"""Tests for models.py."""

from claudre.models import (
    ClaudeState,
    TmuxPane,
    VcsStatus,
    WindowState,
    WindowAdded,
    WindowRemoved,
    WindowStateChanged,
    SummaryUpdated,
)


def test_claude_state_values():
    assert ClaudeState.UNKNOWN.value == "unknown"
    assert ClaudeState.IDLE.value == "idle"
    assert ClaudeState.WORKING.value == "working"
    assert ClaudeState.WAITING.value == "waiting"
    assert ClaudeState.CRASHED.value == "crashed"
    assert ClaudeState.SUSPENDED.value == "suspended"
    assert ClaudeState.OPENING.value == "opening"


def test_tmux_pane_target():
    pane = TmuxPane(
        session="main",
        window_index="3",
        window_name="my-project",
        pane_id="%10",
        pane_pid=12345,
        pane_command="claude",
        pane_path="/home/user/project",
    )
    assert pane.target == "main:3"


def test_window_state_defaults():
    ws = WindowState(pane_id="%1", project_name="test", path="/tmp/test")
    assert ws.state == ClaudeState.UNKNOWN
    assert ws.summary == ""
    assert ws.summary_stale is False
    assert ws.managed is False
    assert ws.session == ""
    assert ws.window_index == ""
    assert isinstance(ws.vcs, VcsStatus)


def test_window_state_target():
    ws = WindowState(
        pane_id="%5",
        project_name="proj",
        path="/tmp/proj",
        session="main",
        window_index="3",
    )
    assert ws.target == "main:3"


def test_window_state_target_empty():
    ws = WindowState(pane_id="%1", project_name="test", path="/tmp")
    assert ws.target == ":"


def test_event_types():
    e1 = WindowAdded(pane_id="%1")
    assert e1.pane_id == "%1"

    e2 = WindowRemoved(pane_id="%2")
    assert e2.pane_id == "%2"

    e3 = WindowStateChanged(pane_id="%3", old=ClaudeState.IDLE, new=ClaudeState.WORKING)
    assert e3.old == ClaudeState.IDLE
    assert e3.new == ClaudeState.WORKING

    e4 = SummaryUpdated(pane_id="%4", summary="doing stuff")
    assert e4.summary == "doing stuff"



"""VCS (git) status detection with async TTL cache."""

from __future__ import annotations

import asyncio
import subprocess
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from pathlib import Path

from claudre.logger import get_logger
from claudre.models import VcsStatus

log = get_logger(__name__)

# Try to import watchdog for inotify-based invalidation
try:
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler
    _WATCHDOG_AVAILABLE = True
except ImportError:
    _WATCHDOG_AVAILABLE = False


@dataclass
class VcsCache:
    ttl: float = 30.0
    _cache: dict[str, tuple[VcsStatus, float]] = field(default_factory=dict)
    _executor: ThreadPoolExecutor = field(default_factory=ThreadPoolExecutor)
    _observer: object | None = field(default=None, init=False, repr=False)

    def __post_init__(self) -> None:
        if _WATCHDOG_AVAILABLE:
            self._observer = Observer()
            self._observer.start()  # type: ignore[union-attr]

    async def get(self, path: str) -> VcsStatus:
        """Return cached VcsStatus, refreshing if stale."""
        entry = self._cache.get(path)
        if entry is not None:
            status, expires = entry
            if time.monotonic() < expires:
                return status

        loop = asyncio.get_event_loop()
        status = await loop.run_in_executor(self._executor, self._fetch_sync, path)
        self._cache[path] = (status, time.monotonic() + self.ttl)

        if _WATCHDOG_AVAILABLE and self._observer is not None:
            self._install_watch(path)

        return status

    def _fetch_sync(self, path: str) -> VcsStatus:
        """Synchronous git status fetch — runs in executor thread."""
        p = Path(path)
        if not (p / ".git").exists():
            return VcsStatus()

        branch = ""
        dirty = False

        try:
            result = subprocess.run(
                ["git", "-C", path, "branch", "--show-current"],
                capture_output=True,
                text=True,
                timeout=5,
            )
            if result.returncode == 0:
                branch = result.stdout.strip()
        except (subprocess.TimeoutExpired, FileNotFoundError):
            pass

        try:
            result = subprocess.run(
                ["git", "-C", path, "status", "--porcelain"],
                capture_output=True,
                text=True,
                timeout=5,
            )
            if result.returncode == 0:
                dirty = bool(result.stdout.strip())
        except (subprocess.TimeoutExpired, FileNotFoundError):
            pass

        return VcsStatus(vcs_type="git", branch=branch, dirty=dirty)

    def invalidate(self, path: str) -> None:
        """Force re-fetch on next access for this path."""
        self._cache.pop(path, None)

    def _install_watch(self, path: str) -> None:
        """Install inotify watches on .git/HEAD and .git/index."""
        if not _WATCHDOG_AVAILABLE or self._observer is None:
            return
        git_dir = Path(path) / ".git"
        if not git_dir.exists():
            return

        cache = self

        class _GitHandler(FileSystemEventHandler):  # type: ignore[misc]
            def on_modified(self, event):
                name = Path(event.src_path).name
                if name in ("HEAD", "index"):
                    cache.invalidate(path)

        try:
            self._observer.schedule(_GitHandler(), str(git_dir), recursive=False)  # type: ignore[union-attr]
        except Exception:
            pass

    def stop(self) -> None:
        if _WATCHDOG_AVAILABLE and self._observer is not None:
            try:
                self._observer.stop()  # type: ignore[union-attr]
                self._observer.join(timeout=2)  # type: ignore[union-attr]
            except Exception:
                pass



"""Async tmux subprocess adapter for claudre v3."""

from __future__ import annotations

import asyncio
import os
from dataclasses import dataclass

from claudre.logger import get_logger
from claudre.models import TmuxPane

log = get_logger(__name__)

_PANE_FMT = (
    "#{session_name}\t#{window_index}\t#{window_name}"
    "\t#{pane_id}\t#{pane_pid}\t#{pane_current_command}\t#{pane_current_path}"
)


class UnmanagedWindowError(Exception):
    """Raised when a mutation is attempted on a non-claudre-managed window."""


@dataclass
class WindowSpec:
    session: str
    template_name: str
    project_name: str
    start_directory: str


class TmuxAdapter:
    MANAGED_OPTION = "@claudre_managed"

    # ------------------------------------------------------------------ #
    # Internal subprocess helper
    # ------------------------------------------------------------------ #

    async def _run(self, args: list[str], check: bool = False) -> tuple[int, str, str]:
        """Run a tmux command asynchronously. Returns (returncode, stdout, stderr)."""
        try:
            proc = await asyncio.create_subprocess_exec(
                *args,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await asyncio.wait_for(proc.communicate(), timeout=5.0)
            rc = proc.returncode or 0
            if check and rc != 0:
                raise RuntimeError(
                    f"tmux command failed (rc={rc}): {' '.join(args)}\n{stderr.decode()}"
                )
            return rc, stdout.decode("utf-8", errors="replace"), stderr.decode("utf-8", errors="replace")
        except asyncio.TimeoutError:
            log.warning("tmux command timed out: %s", " ".join(args))
            return 1, "", "timeout"
        except FileNotFoundError:
            log.error("tmux not found in PATH")
            return 1, "", "not found"

    # ------------------------------------------------------------------ #
    # Queries (no ownership check)
    # ------------------------------------------------------------------ #

    async def list_panes(self) -> list[TmuxPane]:
        """Return all panes across all sessions."""
        rc, stdout, _ = await self._run(["tmux", "list-panes", "-a", "-F", _PANE_FMT])
        if rc != 0:
            return []

        panes: list[TmuxPane] = []
        for line in stdout.strip().splitlines():
            parts = line.split("\t")
            if len(parts) < 7:
                continue
            try:
                panes.append(
                    TmuxPane(
                        session=parts[0],
                        window_index=parts[1],
                        window_name=parts[2],
                        pane_id=parts[3],
                        pane_pid=int(parts[4]),
                        pane_command=parts[5],
                        pane_path=parts[6],
                    )
                )
            except (ValueError, IndexError):
                continue
        return panes

    async def capture_pane(self, pane_id: str, lines: int = 100) -> str:
        """Capture the visible content of a pane."""
        rc, stdout, _ = await self._run(
            ["tmux", "capture-pane", "-p", "-t", pane_id, "-S", f"-{lines}"]
        )
        if rc != 0:
            return ""
        return stdout

    async def get_window_option(self, target: str, option: str) -> str:
        """Get a window option value. Returns empty string if not set."""
        rc, stdout, _ = await self._run(
            ["tmux", "show-options", "-wqv", "-t", target, option]
        )
        if rc != 0:
            return ""
        return stdout.strip()

    async def is_inside_tmux(self) -> bool:
        return "TMUX" in os.environ

    async def current_session(self) -> str:
        rc, stdout, _ = await self._run(
            ["tmux", "display-message", "-p", "#{session_name}"]
        )
        if rc != 0:
            return ""
        return stdout.strip()

    async def current_window_index(self) -> str:
        rc, stdout, _ = await self._run(
            ["tmux", "display-message", "-p", "#{window_index}"]
        )
        if rc != 0:
            return ""
        return stdout.strip()

    # ------------------------------------------------------------------ #
    # Mutations (ownership checked except create + select)
    # ------------------------------------------------------------------ #

    async def create_window(self, spec: WindowSpec) -> TmuxPane:
        """Create a new tmux window. Returns the first pane."""
        rc, stdout, _ = await self._run(
            [
                "tmux", "new-window",
                "-t", spec.session,
                "-c", spec.start_directory,
                "-P", "-F", _PANE_FMT,
            ],
            check=True,
        )
        parts = stdout.strip().split("\t")
        if len(parts) < 7:
            raise RuntimeError(f"Unexpected new-window output: {stdout!r}")
        return TmuxPane(
            session=parts[0],
            window_index=parts[1],
            window_name=parts[2],
            pane_id=parts[3],
            pane_pid=int(parts[4]),
            pane_command=parts[5],
            pane_path=parts[6],
        )

    async def kill_window(self, target: str) -> None:
        await self._assert_managed(target)
        await self._run(["tmux", "kill-window", "-t", target])

    async def send_keys(self, target: str, keys: str, enter: bool = True) -> None:
        cmd = ["tmux", "send-keys", "-t", target, keys]
        if enter:
            cmd.append("Enter")
        await self._run(cmd)

    async def rename_window(self, target: str, name: str) -> None:
        await self._run(["tmux", "rename-window", "-t", target, name])
        # Lock the name
        await self._run(["tmux", "set-option", "-w", "-t", target, "allow-rename", "off"])
        await self._run(["tmux", "set-option", "-w", "-t", target, "automatic-rename", "off"])

    async def select_window(self, target: str) -> None:
        await self._run(["tmux", "select-window", "-t", target])

    async def set_window_option(self, target: str, opt: str, val: str) -> None:
        await self._run(["tmux", "set-option", "-w", "-t", target, opt, val])

    async def set_global_option(self, opt: str, val: str) -> None:
        await self._run(["tmux", "set-option", "-g", opt, val])

    async def split_window(
        self,
        target: str,
        horizontal: bool = True,
        percent: int = 50,
        start_directory: str = "",
    ) -> None:
        """Split a window pane."""
        cmd = ["tmux", "split-window", "-t", target]
        if horizontal:
            cmd.append("-h")
        cmd += ["-p", str(percent)]
        if start_directory:
            cmd += ["-c", start_directory]
        await self._run(cmd)

    async def select_layout(self, target: str, layout: str) -> None:
        """Apply a named tmux layout to a window."""
        await self._run(["tmux", "select-layout", "-t", target, layout])

    async def select_pane(self, pane_id: str) -> None:
        """Focus a specific pane."""
        await self._run(["tmux", "select-pane", "-t", pane_id])

    # ------------------------------------------------------------------ #
    # Ownership helpers
    # ------------------------------------------------------------------ #

    async def is_managed(self, target: str) -> bool:
        """Return True if this window was created by claudre."""
        val = await self.get_window_option(target, self.MANAGED_OPTION)
        return val == "1"

    # Keep private alias for any internal callers
    async def _is_managed(self, target: str) -> bool:
        return await self.is_managed(target)

    async def _assert_managed(self, target: str) -> None:
        if not await self.is_managed(target):
            raise UnmanagedWindowError(target)



"""Structured logging setup for claudre."""

from __future__ import annotations

import logging

_FMT = "%(asctime)s %(levelname)-5s %(name)s: %(message)s"
_DATE_FMT = "%Y-%m-%dT%H:%M:%S"


def setup_logging(log_path: str = "", debug: bool = False) -> None:
    """Configure root logger. Call once at startup."""
    level = logging.DEBUG if debug else logging.INFO
    handlers: list[logging.Handler] = [logging.StreamHandler()]

    if log_path:
        handlers.append(logging.FileHandler(log_path))

    logging.basicConfig(
        level=level,
        format=_FMT,
        datefmt=_DATE_FMT,
        handlers=handlers,
        force=True,
    )


def get_logger(name: str) -> logging.Logger:
    return logging.getLogger(name)



"""Click CLI entry point for claudre v3."""

from __future__ import annotations

import re
from pathlib import Path

import click

from claudre.config import ConfigError, load_config


class DefaultGroup(click.Group):
    """Click group that defaults to 'dashboard' when no subcommand is given."""

    def parse_args(self, ctx, args):
        if not args or (args[0] not in self.commands and args[0] not in ("--help", "-h")):
            args = ["dashboard"] + list(args)
        return super().parse_args(ctx, args)


@click.group(cls=DefaultGroup)
@click.option("--debug", is_flag=True, default=False, help="Enable debug logging.")
@click.option("--log-file", default="", metavar="PATH", help="Write logs to this file.")
@click.pass_context
def main(ctx: click.Context, debug: bool, log_file: str) -> None:
    """claudre — Claude-native tmux session manager."""
    ctx.ensure_object(dict)
    ctx.obj["debug"] = debug
    ctx.obj["log_file"] = log_file

    from claudre.logger import setup_logging
    setup_logging(log_path=log_file, debug=debug)


@main.command()
@click.pass_context
def dashboard(ctx: click.Context) -> None:
    """Launch the persistent dashboard TUI."""
    try:
        config = load_config()
    except ConfigError as e:
        raise click.ClickException(str(e))

    import asyncio
    import sys
    from claudre.tmux_adapter import TmuxAdapter
    tmux = TmuxAdapter()

    async def _switch_or_setup() -> bool:
        """Switch to an existing claudre dashboard if one is open.

        Returns True if we switched (caller should exit), False if a new
        dashboard should be launched (current window renamed to 'claudre').
        """
        if not await tmux.is_inside_tmux():
            return False

        session = await tmux.current_session()
        current_idx = await tmux.current_window_index()
        panes = await tmux.list_panes()
        session_panes = [p for p in panes if p.session == session]

        # Look for any window named "claudre" that isn't the current window
        existing = [
            p for p in session_panes
            if p.window_name == "claudre" and p.window_index != current_idx
        ]
        if existing:
            await tmux.select_window(existing[0].target)
            return True

        # No existing dashboard — rename current window and launch TUI
        current_panes = [p for p in session_panes if p.window_index == current_idx]
        if current_panes:
            try:
                await tmux.rename_window(current_panes[0].target, "claudre")
            except Exception:
                pass
        return False

    if asyncio.run(_switch_or_setup()):
        sys.exit(0)

    from textual.app import App
    from claudre.screens.dashboard import DashboardScreen

    class ClaudreApp(App):
        CSS_PATH = Path(__file__).parent / "css" / "app.tcss"

        def __init__(self, config) -> None:
            super().__init__()
            self._config = config

        def on_mount(self) -> None:
            self.push_screen(DashboardScreen(self._config))

    app = ClaudreApp(config)
    app.run()


@main.command()
@click.argument("template", default="")
@click.pass_context
def new(ctx: click.Context, template: str) -> None:
    """Create a new tmux window from a template (no TUI needed)."""
    import asyncio

    try:
        config = load_config()
    except ConfigError as e:
        raise click.ClickException(str(e))

    template_name = template or config.defaults.template
    cwd = str(Path.cwd())
    project_name = Path(cwd).name

    async def _run():
        from claudre.tmux_adapter import TmuxAdapter, WindowSpec
        from claudre.templates import create_from_template

        tmux = TmuxAdapter()
        if not await tmux.is_inside_tmux():
            raise click.ClickException("Not inside a tmux session")

        session = await tmux.current_session()
        spec = WindowSpec(
            session=session,
            template_name=template_name,
            project_name=project_name,
            start_directory=cwd,
        )
        pane = await create_from_template(tmux, spec, config)
        click.echo(f"Created window '{project_name}' in session '{session}' (pane {pane.pane_id})")

    asyncio.run(_run())


@main.command()
@click.pass_context
def popup(ctx: click.Context) -> None:
    """Quick window-switcher popup (run via: tmux display-popup -E 'claudre popup')."""
    try:
        config = load_config()
    except ConfigError as e:
        raise click.ClickException(str(e))

    from claudre.screens.popup import PopupApp
    app = PopupApp(config)
    app.run()


@main.command()
@click.option("--status-bar", is_flag=True, help="Also add status bar integration")
def setup(status_bar: bool) -> None:
    """Install tmux keybindings (prefix+D, prefix+N, prefix+P) to ~/.tmux.conf."""
    tmux_conf = Path.home() / ".tmux.conf"
    marker = "# claudre-v3"

    lines_to_add = [
        f"\n{marker}",
        'bind-key D run-shell "tmux select-window -t claudre 2>/dev/null || true"',
        'bind-key N run-shell "claudre new"',
        'bind-key P display-popup -E -w 80% -h 40% "claudre popup"',
    ]

    if status_bar:
        lines_to_add.append('set -g status-right "#{@claudre_status} | %H:%M"')

    block = "\n".join(lines_to_add) + "\n"

    if tmux_conf.exists():
        content = tmux_conf.read_text()
        if marker in content:
            content = re.sub(
                r"\n?" + re.escape(marker) + r".*?(?=\n#[^#]|\n\n|\Z)",
                block.rstrip(),
                content,
                flags=re.DOTALL,
            )
            tmux_conf.write_text(content)
            click.echo("Updated claudre bindings in ~/.tmux.conf")
        else:
            tmux_conf.write_text(content.rstrip() + "\n" + block)
            click.echo("Appended claudre bindings to ~/.tmux.conf")
    else:
        tmux_conf.write_text(block)
        click.echo("Created ~/.tmux.conf with claudre bindings")

    click.echo("  prefix+D  → jump to claudre dashboard")
    click.echo("  prefix+N  → create new window")
    click.echo("  prefix+P  → quick window-switcher popup")
    click.echo("\nRun 'tmux source ~/.tmux.conf' to activate.")


@main.command()
def init() -> None:
    """Discover projects in auto_discover_dirs and print results."""
    try:
        config = load_config()
    except ConfigError as e:
        raise click.ClickException(str(e))

    discovered = 0
    for dir_path in config.auto_discover_dirs:
        scan_dir = Path(dir_path).expanduser()
        if not scan_dir.exists():
            click.echo(f"  [skip] {scan_dir} — not found")
            continue
        for d in sorted(scan_dir.iterdir()):
            if d.is_dir() and (d / ".git").exists():
                name = d.name
                if name not in config.projects:
                    click.echo(f"  + {name} ({d})")
                    discovered += 1

    if discovered:
        click.echo(f"\nDiscovered {discovered} project(s).")
        click.echo("Add them to ~/.claudre/config.toml manually or use claudre init --save.")
    else:
        click.echo("No new projects discovered.")


@main.command("list")
def list_cmd() -> None:
    """Print window status table to stdout (no TUI)."""
    import asyncio

    async def _run():
        from claudre.tmux_adapter import TmuxAdapter
        from claudre.state_detector import JournalStateDetector

        tmux = TmuxAdapter()
        detector = JournalStateDetector()
        panes = await tmux.list_panes()

        click.echo(f"{'Pane':<12} {'Window':<12} {'Project':<20} {'State':<12}")
        click.echo("-" * 60)

        for pane in panes:
            state = await detector.detect(pane)
            project = Path(pane.pane_path).name or pane.window_name
            click.echo(
                f"{pane.pane_id:<12} {pane.target:<12} {project:<20} {state.value:<12}"
            )

    asyncio.run(_run())


if __name__ == "__main__":
    main()



"""Toast notification widget for claudre."""

from __future__ import annotations

import asyncio
from collections import deque

from textual.app import App
from textual.widget import Widget
from textual.widgets import Label


class Toast(Widget):
    """Auto-dismissing notification overlay (upper-right)."""

    DEFAULT_CSS = """
    Toast {
        dock: top;
        align: right top;
        margin: 1 2;
        width: auto;
        height: auto;
        background: $surface;
        border: round $accent;
        padding: 0 1;
        layer: overlay;
    }
    """

    def __init__(self, message: str, duration: float = 3.0) -> None:
        super().__init__()
        self._message = message
        self._duration = duration

    def compose(self):
        yield Label(self._message)

    def on_mount(self) -> None:
        self.set_timer(self._duration, self._dismiss)

    def _dismiss(self) -> None:
        self.remove()


class ToastManager:
    """Manages a queue of toast notifications, showing one at a time."""

    def __init__(self, app: App, duration: float = 3.0) -> None:
        self._app = app
        self._duration = duration
        self._queue: deque[str] = deque()
        self._active: Toast | None = None

    def show(self, message: str) -> None:
        """Enqueue a toast message."""
        self._queue.append(message)
        if self._active is None:
            self._show_next()

    def _show_next(self) -> None:
        if not self._queue:
            self._active = None
            return
        msg = self._queue.popleft()
        toast = Toast(msg, duration=self._duration)
        self._active = toast
        self._app.mount(toast)
        self._app.set_timer(self._duration + 0.1, self._show_next)



"""Tests for CLI dashboard switch-or-launch logic."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock, patch
import pytest

from claudre.models import TmuxPane


def make_pane(session: str, window_index: str, window_name: str, pane_id: str) -> TmuxPane:
    return TmuxPane(
        session=session,
        window_index=window_index,
        window_name=window_name,
        pane_id=pane_id,
        pane_pid=100,
        pane_command="bash",
        pane_path="/tmp",
    )


@pytest.mark.asyncio
async def test_switch_to_existing_dashboard():
    """If a 'claudre' window exists in a different index, select it and return True."""
    from claudre.tmux_adapter import TmuxAdapter

    tmux = TmuxAdapter()
    current_pane = make_pane("main", "1", "shell", "%1")
    claudre_pane = make_pane("main", "0", "claudre", "%0")

    with patch.object(tmux, "is_inside_tmux", AsyncMock(return_value=True)), \
         patch.object(tmux, "current_session", AsyncMock(return_value="main")), \
         patch.object(tmux, "current_window_index", AsyncMock(return_value="1")), \
         patch.object(tmux, "list_panes", AsyncMock(return_value=[current_pane, claudre_pane])), \
         patch.object(tmux, "select_window", AsyncMock()) as mock_select, \
         patch.object(tmux, "rename_window", AsyncMock()) as mock_rename:

        switched = await _run_switch_or_setup(tmux)

    assert switched is True
    mock_select.assert_awaited_once_with("main:0")
    mock_rename.assert_not_awaited()


@pytest.mark.asyncio
async def test_no_existing_dashboard_renames_current():
    """If no 'claudre' window exists, rename the current window and return False."""
    from claudre.tmux_adapter import TmuxAdapter

    tmux = TmuxAdapter()
    current_pane = make_pane("main", "2", "shell", "%2")

    with patch.object(tmux, "is_inside_tmux", AsyncMock(return_value=True)), \
         patch.object(tmux, "current_session", AsyncMock(return_value="main")), \
         patch.object(tmux, "current_window_index", AsyncMock(return_value="2")), \
         patch.object(tmux, "list_panes", AsyncMock(return_value=[current_pane])), \
         patch.object(tmux, "select_window", AsyncMock()) as mock_select, \
         patch.object(tmux, "rename_window", AsyncMock()) as mock_rename:

        switched = await _run_switch_or_setup(tmux)

    assert switched is False
    mock_select.assert_not_awaited()
    mock_rename.assert_awaited_once_with("main:2", "claudre")


@pytest.mark.asyncio
async def test_claudre_window_is_current_window_launches_new():
    """If the current window is already named 'claudre', don't switch — launch TUI."""
    from claudre.tmux_adapter import TmuxAdapter

    tmux = TmuxAdapter()
    # Current window IS the claudre window
    claudre_pane = make_pane("main", "0", "claudre", "%0")

    with patch.object(tmux, "is_inside_tmux", AsyncMock(return_value=True)), \
         patch.object(tmux, "current_session", AsyncMock(return_value="main")), \
         patch.object(tmux, "current_window_index", AsyncMock(return_value="0")), \
         patch.object(tmux, "list_panes", AsyncMock(return_value=[claudre_pane])), \
         patch.object(tmux, "select_window", AsyncMock()) as mock_select, \
         patch.object(tmux, "rename_window", AsyncMock()):

        switched = await _run_switch_or_setup(tmux)

    # Window index matches — should not switch, should rename (already named claudre, harmless)
    assert switched is False
    mock_select.assert_not_awaited()


@pytest.mark.asyncio
async def test_not_inside_tmux_returns_false():
    """Outside tmux, always return False (launch normally)."""
    from claudre.tmux_adapter import TmuxAdapter

    tmux = TmuxAdapter()
    with patch.object(tmux, "is_inside_tmux", AsyncMock(return_value=False)):
        switched = await _run_switch_or_setup(tmux)

    assert switched is False


# ---------------------------------------------------------------------------
# Helper — extract the _switch_or_setup logic without Click wiring
# ---------------------------------------------------------------------------

async def _run_switch_or_setup(tmux) -> bool:
    """Replicate the _switch_or_setup coroutine from cli.dashboard()."""
    if not await tmux.is_inside_tmux():
        return False

    session = await tmux.current_session()
    current_idx = await tmux.current_window_index()
    panes = await tmux.list_panes()
    session_panes = [p for p in panes if p.session == session]

    existing = [
        p for p in session_panes
        if p.window_name == "claudre" and p.window_index != current_idx
    ]
    if existing:
        await tmux.select_window(existing[0].target)
        return True

    current_panes = [p for p in session_panes if p.window_index == current_idx]
    if current_panes:
        try:
            await tmux.rename_window(current_panes[0].target, "claudre")
        except Exception:
            pass
    return False



"""Detail panel showing full info for the selected window."""

from __future__ import annotations

import time
from datetime import datetime

from textual.widgets import Static

from claudre.models import ClaudeState, WindowState


class DetailPanel(Static):
    """Panel showing details of the selected tmux window."""

    def update_window(self, ws: WindowState | None) -> None:
        if ws is None:
            self.update("[dim]No window selected[/]")
            return

        lines: list[str] = [f"[bold]{ws.project_name}[/]", f"  Path: {ws.path}"]

        # State
        state_map = {
            ClaudeState.WORKING: "[bold green]WORKING[/]",
            ClaudeState.WAITING: "[bold yellow]WAITING[/]",
            ClaudeState.CRASHED: "[bold red]CRASHED[/]",
            ClaudeState.SUSPENDED: "[dim]SUSPENDED[/]",
            ClaudeState.IDLE: "[dim]idle[/]",
            ClaudeState.OPENING: "[cyan]OPENING[/]",
            ClaudeState.UNKNOWN: "[dim]unknown[/]",
        }
        lines.append(f"  State: {state_map.get(ws.state, str(ws.state))}")

        # VCS
        if ws.vcs.vcs_type:
            branch = ws.vcs.branch or "(detached)"
            dirty = " [bold red]*dirty*[/]" if ws.vcs.dirty else ""
            lines.append(f"  Branch: {branch}{dirty}")

        # Pane info
        lines.append(f"  Pane: {ws.pane_id}")
        managed = "[green]yes[/]" if ws.managed else "[dim]no[/]"
        lines.append(f"  Managed: {managed}")

        # Summary
        lines.append("")
        if ws.summary:
            stale = " [dim](stale)[/]" if ws.summary_stale else ""
            lines.append(f"[bold]Summary[/]{stale}")
            lines.append(f"  {ws.summary}")

            if ws.summary_updated_at:
                age = int(time.time() - ws.summary_updated_at.timestamp())
                lines.append(f"  [dim]Updated {age}s ago[/]")
        else:
            lines.append("[dim]No summary yet (press u to generate)[/]")

        self.update("\n".join(lines))



"""Tests for vcs.py."""

from __future__ import annotations

import subprocess
from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest

from claudre.models import VcsStatus
from claudre.vcs import VcsCache


@pytest.fixture
def cache():
    return VcsCache(ttl=30.0)


@pytest.mark.asyncio
async def test_get_non_git_dir(cache, tmp_path):
    status = await cache.get(str(tmp_path))
    assert status.vcs_type is None
    assert status.branch == ""
    assert status.dirty is False


@pytest.mark.asyncio
async def test_get_git_dir(cache, tmp_path):
    (tmp_path / ".git").mkdir()
    with patch("subprocess.run") as mock_run:
        mock_run.side_effect = [
            MagicMock(returncode=0, stdout="main\n"),
            MagicMock(returncode=0, stdout="M  file.py\n"),
        ]
        status = await cache.get(str(tmp_path))
    assert status.vcs_type == "git"
    assert status.branch == "main"
    assert status.dirty is True


@pytest.mark.asyncio
async def test_cache_returns_cached(cache, tmp_path):
    (tmp_path / ".git").mkdir()
    with patch("subprocess.run") as mock_run:
        mock_run.side_effect = [
            MagicMock(returncode=0, stdout="main\n"),
            MagicMock(returncode=0, stdout=""),
        ]
        status1 = await cache.get(str(tmp_path))
        status2 = await cache.get(str(tmp_path))
    # subprocess.run should only be called twice (for the first fetch)
    assert mock_run.call_count == 2
    assert status1.branch == status2.branch == "main"


@pytest.mark.asyncio
async def test_invalidate_forces_refresh(cache, tmp_path):
    (tmp_path / ".git").mkdir()
    with patch("subprocess.run") as mock_run:
        mock_run.side_effect = [
            MagicMock(returncode=0, stdout="main\n"),
            MagicMock(returncode=0, stdout=""),
            MagicMock(returncode=0, stdout="feature\n"),
            MagicMock(returncode=0, stdout=""),
        ]
        await cache.get(str(tmp_path))
        cache.invalidate(str(tmp_path))
        status = await cache.get(str(tmp_path))
    assert status.branch == "feature"


@pytest.mark.asyncio
async def test_git_timeout_handled(cache, tmp_path):
    (tmp_path / ".git").mkdir()
    with patch("subprocess.run", side_effect=subprocess.TimeoutExpired(cmd="git", timeout=5)):
        status = await cache.get(str(tmp_path))
    assert status.branch == ""
    assert status.dirty is False



"""Tests for registry.py."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from claudre.config import ClaudreConfig
from claudre.models import (
    ClaudeState,
    TmuxPane,
    VcsStatus,
    WindowAdded,
    WindowRemoved,
    WindowState,
    WindowStateChanged,
)
from claudre.registry import SessionRegistry


def make_pane(pane_id: str, path: str = "/tmp", command: str = "claude") -> TmuxPane:
    return TmuxPane(
        session="main",
        window_index="0",
        window_name=pane_id,
        pane_id=pane_id,
        pane_pid=100,
        pane_command=command,
        pane_path=path,
    )


def make_registry() -> tuple[SessionRegistry, MagicMock, MagicMock, MagicMock, MagicMock]:
    config = ClaudreConfig(ai_summaries=False)
    tmux = MagicMock()
    tmux.list_panes = AsyncMock(return_value=[])
    tmux.current_session = AsyncMock(return_value="main")
    tmux._is_managed = AsyncMock(return_value=False)
    detector = MagicMock()
    detector.detect = AsyncMock(return_value=ClaudeState.IDLE)
    vcs = MagicMock()
    vcs.get = AsyncMock(return_value=VcsStatus())
    vcs.stop = MagicMock()
    summary = MagicMock()
    summary.should_update = AsyncMock(return_value=False)

    reg = SessionRegistry(
        config=config,
        tmux=tmux,
        detector=detector,
        vcs=vcs,
        summary=summary,
    )
    return reg, tmux, detector, vcs, summary


@pytest.mark.asyncio
async def test_window_added_on_new_pane():
    reg, tmux, *_ = make_registry()
    events = []
    reg.subscribe(events.append)

    pane = make_pane("%1", path="/tmp/proj")
    tmux.list_panes = AsyncMock(return_value=[pane])

    await reg._refresh_once()

    assert any(isinstance(e, WindowAdded) for e in events)
    assert "%1" in reg.windows


@pytest.mark.asyncio
async def test_window_removed_on_gone_pane():
    reg, tmux, *_ = make_registry()

    pane = make_pane("%1", path="/tmp/proj")
    tmux.list_panes = AsyncMock(return_value=[pane])
    await reg._refresh_once()  # Add the pane

    events = []
    reg.subscribe(events.append)

    tmux.list_panes = AsyncMock(return_value=[])
    await reg._refresh_once()  # Remove the pane

    assert any(isinstance(e, WindowRemoved) for e in events)
    assert "%1" not in reg.windows


@pytest.mark.asyncio
async def test_state_change_emits_event():
    reg, tmux, detector, *_ = make_registry()
    events = []
    reg.subscribe(events.append)

    pane = make_pane("%1")
    tmux.list_panes = AsyncMock(return_value=[pane])
    detector.detect = AsyncMock(return_value=ClaudeState.IDLE)
    await reg._refresh_once()

    # Change state
    detector.detect = AsyncMock(return_value=ClaudeState.WORKING)
    await reg._refresh_once()

    state_changes = [e for e in events if isinstance(e, WindowStateChanged)]
    assert any(e.new == ClaudeState.WORKING for e in state_changes)


def test_subscribe_and_emit():
    reg, *_ = make_registry()
    received = []
    reg.subscribe(received.append)

    event = WindowAdded(pane_id="%99")
    reg._emit(event)
    assert event in received


def test_request_summary_marks_stale():
    reg, *_ = make_registry()
    ws = WindowState(pane_id="%1", project_name="test", path="/tmp")
    reg._windows["%1"] = ws

    reg.request_summary("%1")
    assert ws.summary_stale is True



"""AI summarization engine using Claude Haiku."""

from __future__ import annotations

import asyncio
import json
import os
import time
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from claudre.config import ClaudreConfig
from claudre.logger import get_logger
from claudre.models import ClaudeState, WindowState

if TYPE_CHECKING:
    from claudre.state_detector import JournalStateDetector

log = get_logger(__name__)

SYSTEM_PROMPT = """
You are a concise status reporter for a terminal multiplexer. Given terminal output
and optional conversation history, write 1-2 sentences describing what is currently
happening in this terminal window. Focus on the active task, not tool names.
Be specific about what is being built, fixed, or investigated. If Claude is waiting
for user input, say so and describe what decision or feedback is needed.
""".strip()


@dataclass
class SummaryRequest:
    pane_id: str
    terminal_capture: str
    jsonl_context: list[dict] | None = None  # last 3-5 JSONL messages


class SummaryEngine:
    def __init__(
        self,
        config: ClaudreConfig,
        tmux: object,
        detector: "JournalStateDetector | None" = None,
    ) -> None:
        self._config = config
        self._tmux = tmux
        self._detector = detector
        self._queue: asyncio.Queue[str] = asyncio.Queue()
        self._semaphore = asyncio.Semaphore(config.summary_concurrency)
        self._enabled = config.ai_summaries
        self._client: object | None = None

        if self._enabled:
            api_key = os.environ.get("ANTHROPIC_API_KEY", "")
            if not api_key:
                log.warning(
                    "ANTHROPIC_API_KEY not set — AI summaries disabled"
                )
                self._enabled = False
            else:
                try:
                    import anthropic
                    self._client = anthropic.AsyncAnthropic(api_key=api_key)
                except ImportError:
                    log.warning("anthropic package not installed — AI summaries disabled")
                    self._enabled = False

    @property
    def enabled(self) -> bool:
        return self._enabled

    def update_config(self, config: ClaudreConfig) -> None:
        """Hot-reload config values without restarting the engine."""
        self._config = config
        # Update semaphore capacity if concurrency changed
        self._semaphore = asyncio.Semaphore(config.summary_concurrency)

    async def should_update(self, ws: WindowState) -> bool:
        """True if summary should be refreshed for this window."""
        if not self._enabled:
            return False
        if ws.summary_stale:
            return True
        if ws.summary_updated_at is None:
            return True
        age = time.time() - ws.summary_updated_at.timestamp()
        if ws.state == ClaudeState.WORKING and age > self._config.summary_interval:
            return True
        return False

    async def request_update(self, ws: WindowState, registry: object) -> None:
        """Enqueue a summary refresh for this window."""
        if not self._enabled:
            return
        ws.summary_stale = True
        asyncio.ensure_future(self._update_task(ws, registry))

    async def _update_task(self, ws: WindowState, registry: object) -> None:
        async with self._semaphore:
            try:
                req = await self._collect_input(ws.pane_id, ws.path, ws.state)
                summary = await self._summarize(req)
                if summary:
                    ws.summary = summary
                    ws.summary_updated_at = datetime.now()
                    ws.summary_stale = False
                    from claudre.models import SummaryUpdated
                    registry._emit(SummaryUpdated(pane_id=ws.pane_id, summary=summary))
            except Exception as e:
                log.warning("Summary update failed for %s: %s", ws.pane_id, e)

    async def _summarize(self, req: SummaryRequest) -> str:
        """Call the Anthropic API to generate a summary. Returns '' on error."""
        if not self._enabled or self._client is None:
            return ""

        user_parts = [f"Terminal output:\n```\n{req.terminal_capture[-3000:]}\n```"]

        if req.jsonl_context:
            ctx_lines = []
            for msg in req.jsonl_context[-5:]:
                role = msg.get("role", "?")
                content = msg.get("content", "")
                if isinstance(content, list):
                    text = " ".join(
                        c.get("text", "") for c in content
                        if isinstance(c, dict) and c.get("type") == "text"
                    )
                else:
                    text = str(content)
                ctx_lines.append(f"{role}: {text[:500]}")
            if ctx_lines:
                user_parts.append("Recent conversation:\n" + "\n".join(ctx_lines))

        user_content = "\n\n".join(user_parts)

        try:
            import anthropic
            response = await self._client.messages.create(  # type: ignore[union-attr]
                model=self._config.summary_model,
                max_tokens=150,
                system=SYSTEM_PROMPT,
                messages=[{"role": "user", "content": user_content}],
            )
            return response.content[0].text.strip()
        except Exception as e:
            log.warning("Anthropic API error: %s", e)
            return ""

    async def _collect_input(
        self, pane_id: str, path: str, state: ClaudeState
    ) -> SummaryRequest:
        """Gather terminal capture and JSONL context."""
        from claudre.tmux_adapter import TmuxAdapter
        tmux: TmuxAdapter = self._tmux  # type: ignore[assignment]
        capture = await tmux.capture_pane(pane_id, lines=100)

        jsonl_ctx: list[dict] | None = None
        if state in (ClaudeState.WORKING, ClaudeState.WAITING):
            jsonl_ctx = self._read_jsonl_context(path)

        return SummaryRequest(
            pane_id=pane_id,
            terminal_capture=capture,
            jsonl_context=jsonl_ctx,
        )

    def _read_jsonl_context(self, path: str) -> list[dict] | None:
        """Read the last 5 JSONL messages for a project path.

        Uses the detector's shared path cache when available.
        """
        # Use shared detector cache if we have one
        if self._detector is not None:
            latest = self._detector.get_jsonl_path(path)
        else:
            latest = self._scan_for_jsonl(path)

        if latest is None:
            return None

        messages: list[dict] = []
        try:
            file_size = latest.stat().st_size
            with open(latest, "rb") as f:
                f.seek(max(0, file_size - 50 * 1024))
                data_bytes = f.read()
            lines = data_bytes.decode("utf-8", errors="replace").splitlines()
            for line in reversed(lines):
                line = line.strip()
                if not line:
                    continue
                try:
                    msg = json.loads(line)
                    if msg.get("role") in ("assistant", "user"):
                        messages.insert(0, msg)
                    if len(messages) >= 5:
                        break
                except json.JSONDecodeError:
                    continue
        except OSError:
            return None

        return messages if messages else None

    def _scan_for_jsonl(self, path: str) -> Path | None:
        """Fallback: scan ~/.claude/projects/ without a shared cache."""
        from claudre.state_detector import CLAUDE_DIR
        if not CLAUDE_DIR.exists():
            return None

        for d in CLAUDE_DIR.iterdir():
            if not d.is_dir():
                continue
            for jf in sorted(d.glob("*.jsonl")):
                try:
                    with open(jf) as f:
                        first_line = f.readline().strip()
                    if not first_line:
                        continue
                    data = json.loads(first_line)
                    if data.get("cwd") == path:
                        jsonl_files = list(d.glob("*.jsonl"))
                        if jsonl_files:
                            return max(jsonl_files, key=lambda f: f.stat().st_mtime)
                except (json.JSONDecodeError, OSError):
                    continue
        return None



"""Main dashboard screen for claudre v3."""

from __future__ import annotations

import asyncio
from pathlib import Path

from textual import on, work
from textual.screen import Screen
from textual.containers import Horizontal
from textual.widgets import Footer, Input, Static

from claudre.config import ClaudreConfig
from claudre.logger import get_logger
from claudre.models import (
    RegistryEvent,
    SummaryUpdated,
    WindowAdded,
    WindowRemoved,
    WindowState,
    WindowStateChanged,
)
from claudre.registry import SessionRegistry
from claudre.state_detector import JournalStateDetector
from claudre.summary_engine import SummaryEngine
from claudre.tmux_adapter import TmuxAdapter
from claudre.vcs import VcsCache
from claudre.widgets.detail_panel import DetailPanel
from claudre.widgets.toast import ToastManager
from claudre.widgets.window_table import WindowTable

log = get_logger(__name__)


class DashboardScreen(Screen):
    BINDINGS = [
        ("enter", "jump", "Jump"),
        ("n", "new_window", "New"),
        ("x", "close_window", "Close"),
        ("s", "send_message", "Send"),
        ("r", "run_command", "Run"),
        ("u", "update_summary", "Summary"),
        ("shift+r", "force_refresh", "Refresh"),
        ("/", "filter", "Filter"),
        ("?", "help", "Help"),
        ("q", "quit_app", "Quit"),
    ]

    CSS_PATH = Path(__file__).parents[1] / "css" / "app.tcss"

    def __init__(self, config: ClaudreConfig) -> None:
        super().__init__()
        self._config = config
        self._tmux = TmuxAdapter()
        self._vcs = VcsCache(ttl=config.vcs_cache_ttl)
        self._detector = JournalStateDetector()
        self._summary = SummaryEngine(config, self._tmux, detector=self._detector)
        self._registry = SessionRegistry(
            config=config,
            tmux=self._tmux,
            detector=self._detector,
            vcs=self._vcs,
            summary=self._summary,
        )
        self._toast: ToastManager | None = None
        self._filter_visible = False

    def compose(self):
        yield Static(" claudre ", id="header")
        with Horizontal(id="body"):
            yield WindowTable(id="window-table")
            yield DetailPanel(id="detail-panel")
        yield Footer()

    def on_mount(self) -> None:
        self._toast = ToastManager(self.app)
        self._registry.subscribe(self._on_registry_event)
        self._start_registry()

    @work(exclusive=True)
    async def _start_registry(self) -> None:
        await self._registry.start()

    def on_unmount(self) -> None:
        asyncio.ensure_future(self._registry.stop())

    # ------------------------------------------------------------------ #
    # Registry event handler (called from background task via call_from_thread)
    # ------------------------------------------------------------------ #

    def _on_registry_event(self, event: RegistryEvent) -> None:
        """Dispatch registry events to UI updates (called on the event loop)."""
        self.call_later(self._apply_event, event)

    def _apply_event(self, event: RegistryEvent) -> None:
        table = self.query_one(WindowTable)

        if isinstance(event, WindowAdded):
            ws = self._registry.windows.get(event.pane_id)
            if ws:
                table.add_window(ws)
        elif isinstance(event, WindowRemoved):
            table.remove_window(event.pane_id)
            self._update_detail(None)
        elif isinstance(event, WindowStateChanged):
            ws = self._registry.windows.get(event.pane_id)
            if ws:
                table.update_window(ws)
                if event.pane_id == table.get_selected_pane_id():
                    self._update_detail(ws)
                # Toast on WORKING → WAITING transition
                from claudre.models import ClaudeState
                if (
                    self._config.notification_on_waiting
                    and event.old == ClaudeState.WORKING
                    and event.new == ClaudeState.WAITING
                    and self._toast
                ):
                    msg = ws.summary or "waiting for input"
                    self._toast.show(f"{ws.project_name}: {msg[:60]}")
        elif isinstance(event, SummaryUpdated):
            ws = self._registry.windows.get(event.pane_id)
            if ws:
                table.update_window(ws)
                if event.pane_id == table.get_selected_pane_id():
                    self._update_detail(ws)

    def _update_detail(self, ws: WindowState | None) -> None:
        detail = self.query_one(DetailPanel)
        if ws is None:
            table = self.query_one(WindowTable)
            pane_id = table.get_selected_pane_id()
            ws = self._registry.windows.get(pane_id or "") if pane_id else None
        detail.update_window(ws)

    @on(WindowTable.RowHighlighted)
    def on_row_highlighted(self, event: WindowTable.RowHighlighted) -> None:
        self._update_detail(None)

    # ------------------------------------------------------------------ #
    # Actions
    # ------------------------------------------------------------------ #

    def action_jump(self) -> None:
        ws = self._selected_window()
        if ws and ws.session and ws.window_index:
            asyncio.ensure_future(self._tmux.select_window(ws.target))

    def action_new_window(self) -> None:
        async def _do_new() -> None:
            session = await self._tmux.current_session()
            self.app.push_screen(
                _new_window_screen(self._config, session),
                self._on_new_window_result,
            )

        asyncio.ensure_future(_do_new())

    def _on_new_window_result(self, result: object) -> None:
        if result is None:
            return
        from claudre.templates import create_from_template
        from claudre.tmux_adapter import WindowSpec

        async def _create() -> None:
            try:
                spec = WindowSpec(
                    session=result.session if hasattr(result, "session") else "",
                    template_name=result.template_name,
                    project_name=result.project_name,
                    start_directory=result.start_directory,
                )
                # Use current session if not set
                if not spec.session:
                    spec.session = await self._tmux.current_session()
                await create_from_template(self._tmux, spec, self._config)
                if self._toast:
                    self._toast.show(f"Created window: {result.project_name}")
            except Exception as e:
                log.error("Failed to create window: %s", e)
                if self._toast:
                    self._toast.show(f"Error: {e}")

        asyncio.ensure_future(_create())

    def action_close_window(self) -> None:
        ws = self._selected_window()
        if not ws:
            return

        from claudre.screens.confirm import ConfirmScreen

        def on_confirm(confirmed: bool) -> None:
            if confirmed and ws:
                target = ws.pane_id.rsplit(".", 1)[0] if "." in ws.pane_id else ws.pane_id
                asyncio.ensure_future(self._tmux.kill_window(target))

        self.app.push_screen(
            ConfirmScreen(f"Close window '{ws.project_name}'?"), on_confirm
        )

    def action_send_message(self) -> None:
        ws = self._selected_window()
        if not ws:
            return
        from claudre.screens.send_message import SendMessageScreen

        def on_send(text: str | None) -> None:
            if text and ws:
                asyncio.ensure_future(self._tmux.send_keys(ws.pane_id, text))
                # Queue summary update with delay
                async def _delayed_summary() -> None:
                    await asyncio.sleep(5)
                    self._registry.request_summary(ws.pane_id)

                asyncio.ensure_future(_delayed_summary())

        self.app.push_screen(SendMessageScreen(ws.project_name), on_send)

    def action_run_command(self) -> None:
        ws = self._selected_window()
        if not ws:
            return
        proj_cfg = self._config.projects.get(ws.project_name)
        quick_actions = proj_cfg.quick_actions if proj_cfg else []
        from claudre.screens.run_command import RunCommandScreen

        def on_run(cmd: str | None) -> None:
            if cmd and ws:
                asyncio.ensure_future(self._tmux.send_keys(ws.pane_id, cmd))

        self.app.push_screen(RunCommandScreen(ws.project_name, quick_actions), on_run)

    def action_update_summary(self) -> None:
        ws = self._selected_window()
        if not ws:
            return
        self._registry.request_summary(ws.pane_id)
        if self._toast:
            self._toast.show("Summary refresh queued…")

    def action_force_refresh(self) -> None:
        table = self.query_one(WindowTable)
        table.rebuild(self._registry.windows)

    def action_filter(self) -> None:
        self._toggle_filter()

    def _toggle_filter(self) -> None:
        if self._filter_visible:
            # Hide filter
            try:
                self.query_one("#filter-input").remove()
            except Exception:
                pass
            self._filter_visible = False
            self.query_one(WindowTable).set_filter("")
        else:
            inp = Input(placeholder="Filter…", id="filter-input")
            self.query_one(WindowTable).mount(inp, before=self.query_one(WindowTable))
            inp.focus()
            self._filter_visible = True

    @on(Input.Changed, "#filter-input")
    def on_filter_changed(self, event: Input.Changed) -> None:
        self.query_one(WindowTable).set_filter(event.value)
        # Rebuild visible rows
        self.query_one(WindowTable).rebuild(self._registry.windows)

    @on(Input.Submitted, "#filter-input")
    def on_filter_submitted(self, event: Input.Submitted) -> None:
        self._toggle_filter()
        self.query_one(WindowTable).focus()

    def action_help(self) -> None:
        from claudre.screens.help import HelpScreen
        self.app.push_screen(HelpScreen())

    def action_quit_app(self) -> None:
        self.app.exit()

    # ------------------------------------------------------------------ #
    # Helpers
    # ------------------------------------------------------------------ #

    def _selected_window(self) -> WindowState | None:
        table = self.query_one(WindowTable)
        pane_id = table.get_selected_pane_id()
        if not pane_id:
            return None
        return self._registry.windows.get(pane_id)


def _new_window_screen(config: ClaudreConfig, session: str):
    from claudre.screens.new_window import NewWindowScreen
    return NewWindowScreen(config=config, session=session)



"""Two-step modal for creating a new tmux window."""

from __future__ import annotations

from dataclasses import dataclass

from textual import on
from textual.screen import ModalScreen
from textual.containers import Vertical
from textual.widgets import Input, Label, ListView, ListItem, Static

from claudre.config import ClaudreConfig


@dataclass
class NewWindowResult:
    template_name: str
    project_name: str
    start_directory: str


class NewWindowScreen(ModalScreen[NewWindowResult | None]):
    """Template selector + path/name input."""

    BINDINGS = [("escape", "cancel", "Cancel")]

    def __init__(self, config: ClaudreConfig, session: str) -> None:
        super().__init__()
        self._config = config
        self._session = session
        self._selected_template: str = config.defaults.template

    def compose(self):
        # Build template list: builtins + user-defined
        from claudre.templates import _BUILTIN_TEMPLATES
        template_names = list(_BUILTIN_TEMPLATES.keys())
        for name in self._config.templates:
            if name not in template_names:
                template_names.append(name)

        with Vertical(id="new-window-dialog"):
            yield Label("Select template:")
            yield ListView(
                *[
                    ListItem(Static(name), id=f"tmpl-{name}")
                    for name in template_names
                ],
                id="template-list",
            )
            yield Label("Project directory:")
            yield Input(placeholder="/path/to/project", id="dir-input")
            yield Label("Project name (optional):")
            yield Input(placeholder="leave blank to use dir name", id="name-input")

    def on_mount(self) -> None:
        self.query_one("#template-list", ListView).focus()

    @on(ListView.Selected)
    def on_template_selected(self, event: ListView.Selected) -> None:
        item_id = event.item.id or ""
        if item_id.startswith("tmpl-"):
            self._selected_template = item_id[5:]
            self.query_one("#dir-input", Input).focus()

    @on(Input.Submitted, "#name-input")
    def on_name_submitted(self, event: Input.Submitted) -> None:
        self._submit()

    @on(Input.Submitted, "#dir-input")
    def on_dir_submitted(self, event: Input.Submitted) -> None:
        self.query_one("#name-input", Input).focus()

    def _submit(self) -> None:
        from pathlib import Path
        dir_val = self.query_one("#dir-input", Input).value.strip()
        name_val = self.query_one("#name-input", Input).value.strip()

        if not dir_val:
            return

        start_dir = str(Path(dir_val).expanduser())
        project_name = name_val or Path(start_dir).name

        self.dismiss(
            NewWindowResult(
                template_name=self._selected_template,
                project_name=project_name,
                start_directory=start_dir,
            )
        )

    def action_cancel(self) -> None:
        self.dismiss(None)



"""Modal for sending a message to a Claude pane."""

from __future__ import annotations

from textual import on
from textual.screen import ModalScreen
from textual.containers import Vertical
from textual.widgets import Input, Label


class SendMessageScreen(ModalScreen[str | None]):
    """Single-input modal to send text to a Claude session."""

    BINDINGS = [("escape", "cancel", "Cancel")]

    def __init__(self, project_name: str) -> None:
        super().__init__()
        self._project_name = project_name

    def compose(self):
        with Vertical(id="send-dialog"):
            yield Label(f"Send to [bold]{self._project_name}[/]:")
            yield Input(placeholder="Type message…", id="message-input")

    def on_mount(self) -> None:
        self.query_one("#message-input", Input).focus()

    @on(Input.Submitted)
    def on_submitted(self, event: Input.Submitted) -> None:
        text = event.value.strip()
        self.dismiss(text if text else None)

    def action_cancel(self) -> None:
        self.dismiss(None)



"""Session registry — tracks all tmux windows and emits typed events."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from pathlib import Path

from claudre.config import ClaudreConfig
from claudre.logger import get_logger
from claudre.models import (
    ClaudeState,
    RegistryEvent,
    TmuxPane,
    VcsStatus,
    WindowAdded,
    WindowRemoved,
    WindowState,
    WindowStateChanged,
)
from claudre.state_detector import StateDetector
from claudre.summary_engine import SummaryEngine
from claudre.tmux_adapter import TmuxAdapter
from claudre.vcs import VcsCache

log = get_logger(__name__)


class SessionRegistry:
    def __init__(
        self,
        config: ClaudreConfig,
        tmux: TmuxAdapter,
        detector: StateDetector,
        vcs: VcsCache,
        summary: SummaryEngine,
    ) -> None:
        self._config = config
        self._tmux = tmux
        self._detector = detector
        self._vcs = vcs
        self._summary = summary

        self._windows: dict[str, WindowState] = {}
        self._handlers: list[Callable[[RegistryEvent], None]] = []
        self._refresh_task: asyncio.Task | None = None
        self._config_watch_task: asyncio.Task | None = None
        self._running = False

    # ------------------------------------------------------------------ #
    # Public API
    # ------------------------------------------------------------------ #

    @property
    def windows(self) -> dict[str, WindowState]:
        """Read-only view of current window states (keyed by pane_id)."""
        return dict(self._windows)

    def subscribe(self, handler: Callable[[RegistryEvent], None]) -> None:
        self._handlers.append(handler)

    async def start(self) -> None:
        """Begin the refresh loop and optional config watcher."""
        self._running = True
        self._refresh_task = asyncio.ensure_future(self._refresh_loop())
        self._config_watch_task = asyncio.ensure_future(self._config_watch_loop())

    async def stop(self) -> None:
        self._running = False
        if self._refresh_task:
            self._refresh_task.cancel()
        if self._config_watch_task:
            self._config_watch_task.cancel()
        self._vcs.stop()

    def request_summary(self, pane_id: str) -> None:
        """Manually trigger a summary refresh for a pane."""
        ws = self._windows.get(pane_id)
        if ws:
            ws.summary_stale = True

    # ------------------------------------------------------------------ #
    # Refresh loop
    # ------------------------------------------------------------------ #

    async def _refresh_loop(self) -> None:
        while self._running:
            try:
                await self._refresh_once()
            except Exception as e:
                log.warning("Registry refresh error: %s", e)
            await asyncio.sleep(self._config.refresh_interval)

    async def _refresh_once(self) -> None:
        panes = await self._tmux.list_panes()

        # Filter by scope
        current_session = ""
        if self._config.scope == "session":
            current_session = await self._tmux.current_session()

        if self._config.scope == "session" and current_session:
            panes = [p for p in panes if p.session == current_session]

        pane_ids = {p.pane_id for p in panes}
        existing_ids = set(self._windows.keys())

        # Remove gone panes
        for pid in existing_ids - pane_ids:
            del self._windows[pid]
            self._emit(WindowRemoved(pane_id=pid))

        # Add new panes
        for pane in panes:
            if pane.pane_id not in existing_ids:
                ws = WindowState(
                    pane_id=pane.pane_id,
                    project_name=Path(pane.pane_path).name or pane.window_name,
                    path=pane.pane_path,
                    session=pane.session,
                    window_index=pane.window_index,
                    managed=False,
                )
                # Check managed flag once at creation time
                try:
                    ws.managed = await self._tmux.is_managed(pane.target)
                except Exception:
                    ws.managed = False
                self._windows[pane.pane_id] = ws
                self._emit(WindowAdded(pane_id=pane.pane_id))

        # Update all panes concurrently
        await asyncio.gather(
            *[self._update_window(p) for p in panes],
            return_exceptions=True,
        )

        # Check if summaries need updating
        for ws in list(self._windows.values()):
            if await self._summary.should_update(ws):
                await self._summary.request_update(ws, self)

        # Status bar integration
        if self._config.status_bar_integration:
            await self._update_status_bar()

    async def _update_window(self, pane: TmuxPane) -> None:
        ws = self._windows.get(pane.pane_id)
        if ws is None:
            return

        # Keep metadata in sync
        ws.project_name = pane.window_name or Path(pane.pane_path).name or ws.project_name
        ws.path = pane.pane_path
        ws.session = pane.session
        ws.window_index = pane.window_index

        # Detect state
        try:
            new_state = await self._detector.detect(pane)
        except Exception as e:
            log.debug("State detection error for %s: %s", pane.pane_id, e)
            new_state = ClaudeState.UNKNOWN

        if new_state != ws.state:
            old_state = ws.state
            ws.state = new_state
            ws.summary_stale = True
            self._emit(WindowStateChanged(
                pane_id=pane.pane_id,
                old=old_state,
                new=new_state,
            ))

        # VCS
        try:
            ws.vcs = await self._vcs.get(pane.pane_path)
        except Exception:
            pass

    async def _update_status_bar(self) -> None:
        """Update the tmux status bar with a summary of window states."""
        windows = list(self._windows.values())
        crashed = [w for w in windows if w.state == ClaudeState.CRASHED]
        waiting = [w for w in windows if w.state == ClaudeState.WAITING]
        working = [w for w in windows if w.state == ClaudeState.WORKING]

        if crashed:
            urgent = crashed[0]
            status = f"● {urgent.project_name}: CRASHED"
        elif waiting:
            urgent = waiting[0]
            status = f"● {urgent.project_name}: WAITING"
        elif working:
            status = f"● {len(working)} working"
        else:
            status = ""

        if status:
            try:
                await self._tmux.set_global_option("@claudre_status", status)
            except Exception:
                pass

    # ------------------------------------------------------------------ #
    # Config watcher
    # ------------------------------------------------------------------ #

    async def _config_watch_loop(self) -> None:
        """Watch ~/.claudre/config.toml and reload on change."""
        try:
            from watchdog.observers import Observer
            from watchdog.events import FileSystemEventHandler
        except ImportError:
            return

        config_path = Path.home() / ".claudre" / "config.toml"
        if not config_path.parent.exists():
            return

        registry = self
        changed = asyncio.Event()

        class _Handler(FileSystemEventHandler):  # type: ignore[misc]
            def on_modified(self, event):
                if Path(event.src_path) == config_path:
                    changed.set()

        observer = Observer()
        observer.schedule(_Handler(), str(config_path.parent), recursive=False)
        observer.start()

        try:
            while self._running:
                await asyncio.sleep(1)
                if changed.is_set():
                    changed.clear()
                    try:
                        from claudre.config import load_config
                        new_cfg = load_config()
                        registry._config = new_cfg
                        registry._summary.update_config(new_cfg)
                        log.info("Config reloaded from %s", config_path)
                    except Exception as e:
                        log.warning("Config reload failed: %s", e)
        finally:
            observer.stop()
            observer.join(timeout=2)

    # ------------------------------------------------------------------ #
    # Event emission
    # ------------------------------------------------------------------ #

    def _emit(self, event: RegistryEvent) -> None:
        for handler in self._handlers:
            try:
                handler(event)
            except Exception as e:
                log.warning("Event handler error: %s", e)



"""Load and validate TOML configuration for claudre v3."""

from __future__ import annotations

import tomllib
import warnings
from pathlib import Path

from pydantic import BaseModel, ValidationError

CONFIG_DIR = Path.home() / ".claudre"
CONFIG_PATH = CONFIG_DIR / "config.toml"


class ConfigError(Exception):
    pass


class TemplateConfig(BaseModel):
    layout: str = "even-horizontal"
    pane_commands: list[str] = ["claude", "bash"]
    pane_sizes: list[int] = [50, 50]
    rename_to_project: bool = True


class DefaultsConfig(BaseModel):
    claude_command: str = "claude"
    skip_permissions: bool = False
    template: str = "claude+terminal"
    model: str = ""
    extra_args: str = ""


class ProjectConfig(BaseModel):
    path: str
    template: str | None = None
    model: str | None = None
    extra_args: str | None = None
    quick_actions: list[str] = []


class ClaudreConfig(BaseModel):
    refresh_interval: float = 2.0
    vcs_cache_ttl: float = 30.0
    summary_interval: float = 30.0
    summary_concurrency: int = 3
    summary_model: str = "claude-haiku-4-5-20251001"
    ai_summaries: bool = True
    notification_on_waiting: bool = True
    status_bar_integration: bool = False
    debug_log: str = ""
    confirm_new_window: bool = True
    auto_discover_dirs: list[str] = []
    scope: str = "all"  # "all" | "session"
    defaults: DefaultsConfig = DefaultsConfig()
    templates: dict[str, TemplateConfig] = {}
    projects: dict[str, ProjectConfig] = {}


def _migrate_raw(raw: dict) -> dict:
    """Apply v2 → v3 migration transforms, emitting warnings."""
    # Top-level dangerously_skip_permissions → defaults.skip_permissions
    if "dangerously_skip_permissions" in raw:
        warnings.warn(
            "Config: 'dangerously_skip_permissions' is deprecated; "
            "use 'defaults.skip_permissions'",
            DeprecationWarning,
            stacklevel=3,
        )
        raw.setdefault("defaults", {})
        raw["defaults"].setdefault("skip_permissions", raw.pop("dangerously_skip_permissions"))

    # defaults.dangerously_skip_permissions → defaults.skip_permissions
    defaults = raw.get("defaults", {})
    if "dangerously_skip_permissions" in defaults:
        warnings.warn(
            "Config: 'defaults.dangerously_skip_permissions' is deprecated; "
            "use 'defaults.skip_permissions'",
            DeprecationWarning,
            stacklevel=3,
        )
        defaults.setdefault("skip_permissions", defaults.pop("dangerously_skip_permissions"))

    # Per-project layout → template
    for name, proj in raw.get("projects", {}).items():
        if "layout" in proj:
            warnings.warn(
                f"Config: projects.{name}.layout is deprecated; use template",
                DeprecationWarning,
                stacklevel=3,
            )
            proj.setdefault("template", proj.pop("layout"))

    return raw


def load_config() -> ClaudreConfig:
    """Load config from ~/.claudre/config.toml, returning defaults if absent."""
    if not CONFIG_PATH.exists():
        return ClaudreConfig()

    try:
        raw = tomllib.loads(CONFIG_PATH.read_text())
    except tomllib.TOMLDecodeError as e:
        raise ConfigError(f"Config parse error: {e}") from e

    raw = _migrate_raw(raw)

    try:
        return ClaudreConfig.model_validate(raw)
    except ValidationError as e:
        raise ConfigError(f"Config validation error:\n{e}") from e



"""Detect Claude's running state by reading JSONL session files."""

from __future__ import annotations

import json
import time
from pathlib import Path
from typing import Protocol

from claudre.logger import get_logger
from claudre.models import ClaudeState, TmuxPane

log = get_logger(__name__)

CLAUDE_DIR = Path.home() / ".claude" / "projects"
TAIL_BYTES = 100 * 1024  # 100KB tail read
_PATH_CACHE_TTL = 60.0   # seconds before rebuilding path→dir lookup


class StateDetector(Protocol):
    async def detect(self, pane: TmuxPane) -> ClaudeState: ...


class JournalStateDetector:
    """Detect Claude state using JSONL session files and psutil CPU usage."""

    def __init__(self) -> None:
        # path → (sanitized_dir, expiry_monotonic)
        self._path_cache: dict[str, tuple[str, float]] = {}

    async def detect(self, pane: TmuxPane) -> ClaudeState:
        """Detect state for the given pane."""
        # Check if the primary process is alive
        alive = self._is_process_alive(pane.pane_pid)

        if not alive:
            # Process is gone — check last JSONL role to distinguish crash vs suspend
            san_dir = self._get_sanitized_dir(pane.pane_path)
            if san_dir is None:
                return ClaudeState.UNKNOWN
            jsonl = self._find_latest_jsonl(san_dir)
            if jsonl is None:
                return ClaudeState.UNKNOWN
            last_role = self._tail_read_last_role(jsonl)
            if last_role == "user":
                return ClaudeState.CRASHED
            if last_role == "assistant":
                return ClaudeState.SUSPENDED
            return ClaudeState.UNKNOWN

        # Process is alive — check if claude is the pane command
        if pane.pane_command not in ("claude", "node"):
            # Shell or other command running in pane, not claude
            return ClaudeState.IDLE

        san_dir = self._get_sanitized_dir(pane.pane_path)
        if san_dir is None:
            # No JSONL yet — fresh session starting up
            return ClaudeState.OPENING

        jsonl = self._find_latest_jsonl(san_dir)
        if jsonl is None:
            return ClaudeState.OPENING

        try:
            mtime = jsonl.stat().st_mtime
        except OSError:
            return ClaudeState.WAITING

        age = time.time() - mtime

        # Use CPU percent as a more reliable WORKING indicator
        cpu = self._get_cpu_percent(pane.pane_pid)
        if cpu > 5.0 or age < 10.0:
            return ClaudeState.WORKING

        # File is older and CPU is low — check last role
        last_role = self._tail_read_last_role(jsonl)
        if last_role == "assistant":
            return ClaudeState.WAITING
        # last role is "user" or unknown → still processing
        return ClaudeState.WORKING

    def _is_process_alive(self, pid: int) -> bool:
        if pid <= 0:
            return False
        try:
            import psutil
            return psutil.pid_exists(pid)
        except ImportError:
            # Fall back to /proc check
            return Path(f"/proc/{pid}").exists()

    def _get_cpu_percent(self, pid: int) -> float:
        try:
            import psutil
            proc = psutil.Process(pid)
            return proc.cpu_percent(interval=0.0)
        except Exception:
            return 0.0

    def _get_sanitized_dir(self, path: str) -> str | None:
        """Return cached sanitized dir for path, rebuilding if stale."""
        entry = self._path_cache.get(path)
        if entry is not None:
            san_dir, expiry = entry
            if time.monotonic() < expiry:
                return san_dir

        # Need to (re)build
        new_cache = self._build_path_cache()
        now = time.monotonic()
        for p, d in new_cache.items():
            self._path_cache[p] = (d, now + _PATH_CACHE_TTL)

        result = new_cache.get(path)
        if result is None:
            # Mark negative so we don't rebuild every call
            self._path_cache[path] = ("", now + _PATH_CACHE_TTL)
            return None
        return result

    def _build_path_cache(self) -> dict[str, str]:
        """Scan ~/.claude/projects/ and build {real_path: sanitized_dir} lookup."""
        cache: dict[str, str] = {}
        if not CLAUDE_DIR.exists():
            return cache
        for d in CLAUDE_DIR.iterdir():
            if not d.is_dir():
                continue
            jsonl_files = sorted(d.glob("*.jsonl"))
            for jf in jsonl_files:
                try:
                    with open(jf) as f:
                        first_line = f.readline().strip()
                    if not first_line:
                        continue
                    data = json.loads(first_line)
                    cwd = data.get("cwd", "")
                    if cwd:
                        cache[cwd] = d.name
                        break
                except (json.JSONDecodeError, OSError):
                    continue
        return cache

    def get_jsonl_path(self, path: str) -> Path | None:
        """Public: return the latest JSONL file path for a project working dir, or None."""
        san_dir = self._get_sanitized_dir(path)
        if not san_dir:
            return None
        return self._find_latest_jsonl(san_dir)

    def _find_latest_jsonl(self, sanitized_dir: str) -> Path | None:
        """Find the most recently modified JSONL file for a project."""
        if not sanitized_dir:
            return None
        proj_dir = CLAUDE_DIR / sanitized_dir
        if not proj_dir.exists():
            return None
        jsonl_files = list(proj_dir.glob("*.jsonl"))
        if not jsonl_files:
            return None
        return max(jsonl_files, key=lambda f: f.stat().st_mtime)

    def _tail_read_last_role(self, path: Path) -> str | None:
        """Read the last role from a JSONL file using efficient tail-read."""
        try:
            file_size = path.stat().st_size
        except OSError:
            return None
        if file_size == 0:
            return None

        try:
            with open(path, "rb") as f:
                offset = max(0, file_size - TAIL_BYTES)
                f.seek(offset)
                data = f.read()
        except OSError:
            return None

        lines = data.decode("utf-8", errors="replace").strip().splitlines()
        for line in reversed(lines):
            line = line.strip()
            if not line:
                continue
            try:
                msg = json.loads(line)
                role = msg.get("role")
                if role in ("assistant", "user"):
                    return role
            except json.JSONDecodeError:
                continue
        return None



"""Modal for running a quick action or custom command."""

from __future__ import annotations

from textual import on
from textual.screen import ModalScreen
from textual.containers import Vertical
from textual.widgets import Input, Label, ListView, ListItem, Static


class RunCommandScreen(ModalScreen[str | None]):
    """ListView of quick_actions + free-form input."""

    BINDINGS = [("escape", "cancel", "Cancel")]

    def __init__(self, project_name: str, quick_actions: list[str]) -> None:
        super().__init__()
        self._project_name = project_name
        self._quick_actions = quick_actions

    def compose(self):
        with Vertical(id="run-dialog"):
            yield Label(f"Run in [bold]{self._project_name}[/]:")
            if self._quick_actions:
                yield ListView(
                    *[
                        ListItem(Static(action), id=f"action-{i}")
                        for i, action in enumerate(self._quick_actions)
                    ],
                    id="action-list",
                )
            yield Label("[dim]Or enter command:[/]")
            yield Input(placeholder="bash command…", id="cmd-input")

    def on_mount(self) -> None:
        if self._quick_actions:
            self.query_one(ListView).focus()
        else:
            self.query_one("#cmd-input", Input).focus()

    @on(ListView.Selected)
    def on_action_selected(self, event: ListView.Selected) -> None:
        item_id = event.item.id or ""
        if item_id.startswith("action-"):
            idx = int(item_id[7:])
            self.dismiss(self._quick_actions[idx])

    @on(Input.Submitted)
    def on_cmd_submitted(self, event: Input.Submitted) -> None:
        text = event.value.strip()
        self.dismiss(text if text else None)

    def action_cancel(self) -> None:
        self.dismiss(None)



"""Tests for summary_engine.py."""

from __future__ import annotations

import time
from datetime import datetime
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from claudre.config import ClaudreConfig
from claudre.models import ClaudeState, WindowState
from claudre.summary_engine import SummaryEngine, SummaryRequest


def make_config(**kwargs) -> ClaudreConfig:
    return ClaudreConfig(**kwargs)


def make_ws(state: ClaudeState = ClaudeState.WORKING, stale: bool = False) -> WindowState:
    ws = WindowState(pane_id="%1", project_name="test", path="/tmp/test")
    ws.state = state
    ws.summary_stale = stale
    return ws


@pytest.fixture
def engine():
    config = make_config(ai_summaries=False)
    tmux = MagicMock()
    return SummaryEngine(config, tmux)


def test_engine_disabled_without_api_key():
    config = make_config(ai_summaries=True)
    tmux = MagicMock()
    with patch.dict("os.environ", {}, clear=True):
        # Remove ANTHROPIC_API_KEY if present
        import os
        os.environ.pop("ANTHROPIC_API_KEY", None)
        eng = SummaryEngine(config, tmux)
    assert eng.enabled is False


def test_engine_disabled_when_config_off():
    config = make_config(ai_summaries=False)
    tmux = MagicMock()
    eng = SummaryEngine(config, tmux)
    assert eng.enabled is False


@pytest.mark.asyncio
async def test_should_update_disabled(engine):
    ws = make_ws()
    assert await engine.should_update(ws) is False


@pytest.mark.asyncio
async def test_should_update_stale():
    config = make_config(ai_summaries=True, summary_interval=30.0)
    tmux = MagicMock()
    with patch.dict("os.environ", {"ANTHROPIC_API_KEY": "fake"}):
        with patch("anthropic.AsyncAnthropic"):
            eng = SummaryEngine(config, tmux)
    eng._enabled = True

    ws = make_ws(stale=True)
    assert await eng.should_update(ws) is True


@pytest.mark.asyncio
async def test_should_update_no_summary():
    config = make_config(ai_summaries=True)
    tmux = MagicMock()
    eng = SummaryEngine(config, tmux)
    eng._enabled = True

    ws = make_ws()
    ws.summary_updated_at = None
    assert await eng.should_update(ws) is True


@pytest.mark.asyncio
async def test_should_update_working_old():
    config = make_config(ai_summaries=True, summary_interval=30.0)
    tmux = MagicMock()
    eng = SummaryEngine(config, tmux)
    eng._enabled = True

    ws = make_ws(state=ClaudeState.WORKING)
    ws.summary_updated_at = datetime.fromtimestamp(time.time() - 60)
    assert await eng.should_update(ws) is True


@pytest.mark.asyncio
async def test_summarize_returns_empty_on_error(engine):
    engine._enabled = True
    engine._client = MagicMock()
    engine._client.messages = MagicMock()
    engine._client.messages.create = AsyncMock(side_effect=Exception("API error"))

    req = SummaryRequest(pane_id="%1", terminal_capture="hello")
    result = await engine._summarize(req)
    assert result == ""


@pytest.mark.asyncio
async def test_request_update_noop_when_disabled(engine):
    ws = make_ws()
    registry = MagicMock()
    # Should not raise, should be a no-op
    await engine.request_update(ws, registry)
    assert ws.summary_stale is False  # disabled, so stale not set


def test_update_config_changes_interval():
    config1 = make_config(ai_summaries=False, summary_interval=30.0, summary_concurrency=2)
    tmux = MagicMock()
    eng = SummaryEngine(config1, tmux)
    assert eng._config.summary_interval == 30.0

    config2 = make_config(ai_summaries=False, summary_interval=60.0, summary_concurrency=2)
    eng.update_config(config2)
    assert eng._config.summary_interval == 60.0


def test_engine_uses_detector_for_jsonl(tmp_path):
    """SummaryEngine delegates JSONL path resolution to the shared detector."""
    config = make_config(ai_summaries=False)
    tmux = MagicMock()

    detector = MagicMock()
    jsonl = tmp_path / "session.jsonl"
    jsonl.write_text('{"role": "user", "content": "hello"}\n')
    detector.get_jsonl_path = MagicMock(return_value=jsonl)

    eng = SummaryEngine(config, tmux, detector=detector)
    result = eng._read_jsonl_context("/some/path")

    detector.get_jsonl_path.assert_called_once_with("/some/path")
    assert result is not None
    assert result[0]["role"] == "user"



"""Help screen showing all keybindings."""

from __future__ import annotations

from textual.screen import ModalScreen
from textual.containers import Vertical
from textual.widgets import Label, Static

_HELP_TEXT = """\
[bold]claudre v3 — Keybindings[/]

[bold]Navigation[/]
  ↑/↓       Move selection
  Enter     Jump to selected window
  /         Filter windows

[bold]Window Actions[/]
  n         New window (from template)
  x         Close selected window
  u         Update AI summary
  R         Force full refresh

[bold]Send to Claude[/]
  s         Send message to Claude
  r         Run quick action / command

[bold]Display[/]
  ?         Show this help
  q         Quit dashboard
  Escape    Cancel / dismiss modal

[bold]tmux bindings (set via claudre setup)[/]
  prefix+D  Jump to claudre dashboard
  prefix+N  Create new window
"""


class HelpScreen(ModalScreen[None]):
    BINDINGS = [
        ("escape", "dismiss_help", "Close"),
        ("q", "dismiss_help", "Close"),
        ("?", "dismiss_help", "Close"),
    ]

    def compose(self):
        with Vertical(id="help-dialog"):
            yield Static(_HELP_TEXT)

    def action_dismiss_help(self) -> None:
        self.dismiss(None)



"""Quick window-switcher popup for claudre.

Designed to run inside `tmux display-popup -E`. Shows all tmux windows with
their current state. Type to filter, arrow keys to navigate, Enter to jump,
Escape to cancel. On selection the window is focused and the process exits,
which causes tmux to close the popup.
"""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from pathlib import Path

from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.events import Key
from textual.screen import Screen
from textual.widgets import Footer, Input, Label, ListItem, ListView, Static

from claudre.config import ClaudreConfig
from claudre.logger import get_logger
from claudre.models import ClaudeState, TmuxPane
from claudre.state_detector import JournalStateDetector
from claudre.tmux_adapter import TmuxAdapter

log = get_logger(__name__)


# ── Data ─────────────────────────────────────────────────────────────────────

@dataclass
class _Entry:
    pane: TmuxPane
    name: str
    state: ClaudeState


def _state_icon(state: ClaudeState) -> str:
    match state:
        case ClaudeState.WORKING:   return "◌"
        case ClaudeState.WAITING:   return "●"
        case ClaudeState.CRASHED:   return "⚠"
        case ClaudeState.SUSPENDED: return "—"
        case ClaudeState.OPENING:   return "…"
        case _:                     return " "


def _state_markup(state: ClaudeState) -> str:
    match state:
        case ClaudeState.WORKING:   return "[bold green]WORKING[/]"
        case ClaudeState.WAITING:   return "[bold yellow]WAITING[/]"
        case ClaudeState.CRASHED:   return "[bold red]CRASHED[/]"
        case ClaudeState.SUSPENDED: return "[dim]SUSPENDED[/]"
        case ClaudeState.IDLE:      return "[dim]idle[/]"
        case ClaudeState.OPENING:   return "[cyan]OPENING[/]"
        case _:                     return "[dim]—[/]"


# ── List item widget ──────────────────────────────────────────────────────────

class _WindowItem(ListItem):
    """A single row in the popup list."""

    DEFAULT_CSS = """
    _WindowItem {
        height: 1;
        padding: 0 1;
    }
    _WindowItem:focus-within,
    _WindowItem.--highlight {
        background: $accent 30%;
    }
    """

    def __init__(self, entry: _Entry) -> None:
        super().__init__()
        self._entry = entry

    def compose(self) -> ComposeResult:
        icon = _state_icon(self._entry.state)
        state_mu = _state_markup(self._entry.state)
        target = f"{self._entry.pane.target:<10}"
        name = self._entry.name[:28]
        yield Static(
            f"[dim]{target}[/]  {icon} {name:<28}  {state_mu}",
            markup=True,
        )


# ── Screen ────────────────────────────────────────────────────────────────────

class PopupScreen(Screen):
    """Filterable window list for the popup switcher."""

    BINDINGS = [
        Binding("escape", "cancel", "Cancel", show=True),
    ]

    CSS = """
    PopupScreen {
        layout: vertical;
        background: $surface;
    }
    #filter-bar {
        height: 3;
        padding: 0 1;
        border-bottom: solid $accent;
    }
    #filter-input {
        border: none;
        background: transparent;
        width: 1fr;
        height: 1;
        margin: 1 0;
    }
    #window-list {
        height: 1fr;
        border: none;
        padding: 0;
    }
    #empty-label {
        color: $text-muted;
        padding: 1 2;
    }
    """

    def __init__(self, config: ClaudreConfig) -> None:
        super().__init__()
        self._config = config
        self._all_entries: list[_Entry] = []

    def compose(self) -> ComposeResult:
        yield Input(placeholder="type to filter…", id="filter-input")
        yield ListView(id="window-list")
        yield Footer()

    def on_mount(self) -> None:
        self.query_one("#filter-input", Input).focus()
        self.run_worker(self._load_windows(), exclusive=True)

    async def _load_windows(self) -> None:
        tmux = TmuxAdapter()
        detector = JournalStateDetector()
        panes = await tmux.list_panes()

        if self._config.scope == "session":
            session = await tmux.current_session()
            panes = [p for p in panes if p.session == session]

        # One entry per window — take the first pane per (session, window_index)
        seen: set[tuple[str, str]] = set()
        unique: list[TmuxPane] = []
        for p in panes:
            key = (p.session, p.window_index)
            if key not in seen:
                seen.add(key)
                unique.append(p)

        # Skip the claudre dashboard window itself
        current_idx = await tmux.current_window_index()
        current_session = await tmux.current_session()
        unique = [
            p for p in unique
            if not (p.session == current_session and p.window_index == current_idx
                    and p.window_name == "claudre")
        ]

        # Detect states concurrently
        raw = await asyncio.gather(
            *[detector.detect(p) for p in unique],
            return_exceptions=True,
        )

        self._all_entries = [
            _Entry(
                pane=p,
                name=p.window_name or Path(p.pane_path).name or p.pane_id,
                state=s if isinstance(s, ClaudeState) else ClaudeState.UNKNOWN,
            )
            for p, s in zip(unique, raw)
        ]
        self._render_list("")

    def _render_list(self, needle: str) -> None:
        lst = self.query_one("#window-list", ListView)
        lst.clear()
        for entry in self._all_entries:
            if needle and (
                needle not in entry.name.lower()
                and needle not in entry.state.value.lower()
                and needle not in entry.pane.target.lower()
            ):
                continue
            lst.append(_WindowItem(entry))
        # Keep cursor on first visible item
        if lst._nodes:
            lst.index = 0

    # ── Key handling ──────────────────────────────────────────────────────────

    def on_key(self, event: Key) -> None:
        """Route arrow keys and Enter to the list while keeping filter focused."""
        if event.key == "down":
            self._move_selection(1)
            event.prevent_default()
            event.stop()
        elif event.key == "up":
            self._move_selection(-1)
            event.prevent_default()
            event.stop()
        elif event.key == "enter":
            self._select_highlighted()
            event.prevent_default()
            event.stop()

    def _move_selection(self, delta: int) -> None:
        lst = self.query_one("#window-list", ListView)
        count = lst.item_count
        if count == 0:
            return
        current = lst.index if lst.index is not None else -1
        lst.index = max(0, min(count - 1, current + delta))

    def _select_highlighted(self) -> None:
        lst = self.query_one("#window-list", ListView)
        if lst.index is None or lst.item_count == 0:
            return
        item = lst._nodes[lst.index] if lst.index < len(lst._nodes) else None
        if isinstance(item, _WindowItem):
            asyncio.ensure_future(self._switch_and_exit(item._entry))

    # ── Events ────────────────────────────────────────────────────────────────

    @on(Input.Changed, "#filter-input")
    def on_filter_changed(self, event: Input.Changed) -> None:
        self._render_list(event.value.lower())

    @on(ListView.Selected)
    def on_item_selected(self, event: ListView.Selected) -> None:
        if isinstance(event.item, _WindowItem):
            asyncio.ensure_future(self._switch_and_exit(event.item._entry))

    def action_cancel(self) -> None:
        self.app.exit()

    # ── Switch ────────────────────────────────────────────────────────────────

    async def _switch_and_exit(self, entry: _Entry) -> None:
        try:
            tmux = TmuxAdapter()
            await tmux.select_window(entry.pane.target)
        except Exception as e:
            log.warning("Failed to select window %s: %s", entry.pane.target, e)
        finally:
            self.app.exit()


# ── App ───────────────────────────────────────────────────────────────────────

class PopupApp(App):
    """Minimal Textual app hosting the popup switcher screen."""

    CSS = """
    App {
        background: $surface;
    }
    """

    def __init__(self, config: ClaudreConfig) -> None:
        super().__init__()
        self._config = config

    def on_mount(self) -> None:
        self.push_screen(PopupScreen(self._config))



"""Window template creation for claudre v3."""

from __future__ import annotations

from claudre.config import ClaudreConfig, TemplateConfig
from claudre.logger import get_logger
from claudre.models import TmuxPane
from claudre.tmux_adapter import TmuxAdapter, WindowSpec

log = get_logger(__name__)

# Built-in templates
_BUILTIN_TEMPLATES: dict[str, TemplateConfig] = {
    "claude+terminal": TemplateConfig(
        layout="even-horizontal",
        pane_commands=["claude", "bash"],
        pane_sizes=[50, 50],
        rename_to_project=True,
    ),
    "claude+vim+terminal": TemplateConfig(
        layout="main-left",
        pane_commands=["claude", "vim", "bash"],
        pane_sizes=[50, 30, 20],
        rename_to_project=True,
    ),
    "claude": TemplateConfig(
        layout="even-horizontal",
        pane_commands=["claude"],
        pane_sizes=[100],
        rename_to_project=True,
    ),
}


def resolve_template(name: str, config: ClaudreConfig) -> TemplateConfig:
    """Return the TemplateConfig for a given name, checking user config first."""
    if name in config.templates:
        return config.templates[name]
    if name in _BUILTIN_TEMPLATES:
        return _BUILTIN_TEMPLATES[name]
    log.warning("Unknown template %r; falling back to claude+terminal", name)
    return _BUILTIN_TEMPLATES["claude+terminal"]


async def create_from_template(
    tmux: TmuxAdapter,
    spec: WindowSpec,
    config: ClaudreConfig,
) -> TmuxPane:
    """Create a tmux window from a template. Returns the primary (claude) pane."""
    template = resolve_template(spec.template_name, config)

    # Create the base window
    primary = await tmux.create_window(spec)
    target = primary.target

    # Mark as managed immediately
    await tmux.set_window_option(target, tmux.MANAGED_OPTION, "1")

    # Rename window
    if template.rename_to_project:
        await tmux.rename_window(target, spec.project_name)

    commands = template.pane_commands
    sizes = template.pane_sizes

    if len(commands) > 1:
        # Apply the named layout first
        await tmux.select_layout(target, template.layout)

        # Split panes for each additional command
        for i in range(1, len(commands)):
            percent = sizes[i] if i < len(sizes) else 50
            await tmux.split_window(
                target,
                horizontal=True,
                percent=percent,
                start_directory=spec.start_directory,
            )

        # Re-apply layout to even things out
        await tmux.select_layout(target, template.layout)

    # Send commands to each pane in the new window
    all_panes = await tmux.list_panes()
    window_panes = [
        p for p in all_panes
        if p.session == primary.session and p.window_index == primary.window_index
    ]

    for i, pane in enumerate(window_panes):
        if i < len(commands):
            cmd = commands[i]
            if config.defaults.skip_permissions and cmd == config.defaults.claude_command:
                cmd += " --dangerously-skip-permissions"
            if config.defaults.model and cmd.startswith(config.defaults.claude_command):
                cmd += f" --model {config.defaults.model}"
            if config.defaults.extra_args and cmd.startswith(config.defaults.claude_command):
                cmd += f" {config.defaults.extra_args}"
            await tmux.send_keys(pane.pane_id, cmd, enter=True)

    # Focus the first (claude) pane
    if window_panes:
        await tmux.select_pane(window_panes[0].pane_id)

    return primary



"""Tests for popup screen logic."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock

import pytest

from claudre.config import ClaudreConfig
from claudre.models import ClaudeState, TmuxPane
from claudre.screens.popup import _Entry, _state_icon, _state_markup, _WindowItem


def make_pane(session: str, window_index: str, window_name: str, pane_id: str) -> TmuxPane:
    return TmuxPane(
        session=session,
        window_index=window_index,
        window_name=window_name,
        pane_id=pane_id,
        pane_pid=100,
        pane_command="bash",
        pane_path="/tmp",
    )


# ── Icon / markup helpers ─────────────────────────────────────────────────────

def test_state_icons_cover_all_states():
    for state in ClaudeState:
        icon = _state_icon(state)
        assert isinstance(icon, str) and len(icon) == 1


def test_state_markup_contains_label():
    assert "WORKING" in _state_markup(ClaudeState.WORKING)
    assert "WAITING" in _state_markup(ClaudeState.WAITING)
    assert "CRASHED" in _state_markup(ClaudeState.CRASHED)
    assert "idle"    in _state_markup(ClaudeState.IDLE)


# ── _load_windows deduplication logic (extracted for unit testing) ────────────

def _deduplicate(panes: list[TmuxPane]) -> list[TmuxPane]:
    """Mirror the deduplication logic from PopupScreen._load_windows."""
    seen: set[tuple[str, str]] = set()
    unique: list[TmuxPane] = []
    for p in panes:
        key = (p.session, p.window_index)
        if key not in seen:
            seen.add(key)
            unique.append(p)
    return unique


def test_deduplication_one_per_window():
    panes = [
        make_pane("main", "0", "shell",  "%0"),
        make_pane("main", "0", "shell",  "%1"),   # second pane in same window
        make_pane("main", "1", "my-api", "%2"),
    ]
    unique = _deduplicate(panes)
    assert len(unique) == 2
    assert unique[0].pane_id == "%0"
    assert unique[1].pane_id == "%2"


def test_deduplication_across_sessions():
    panes = [
        make_pane("main", "0", "shell",  "%0"),
        make_pane("work", "0", "editor", "%1"),
    ]
    unique = _deduplicate(panes)
    assert len(unique) == 2


# ── Filter logic (mirrors PopupScreen._render_list filter) ───────────────────

def _matches(entry: _Entry, needle: str) -> bool:
    if not needle:
        return True
    return (
        needle in entry.name.lower()
        or needle in entry.state.value.lower()
        or needle in entry.pane.target.lower()
    )


def test_filter_by_name():
    entry = _Entry(
        pane=make_pane("main", "1", "my-api", "%1"),
        name="my-api",
        state=ClaudeState.WAITING,
    )
    assert _matches(entry, "api")
    assert _matches(entry, "my-api")
    assert not _matches(entry, "frontend")


def test_filter_by_state():
    entry = _Entry(
        pane=make_pane("main", "2", "frontend", "%2"),
        name="frontend",
        state=ClaudeState.CRASHED,
    )
    assert _matches(entry, "crashed")
    assert not _matches(entry, "waiting")


def test_filter_by_target():
    entry = _Entry(
        pane=make_pane("work", "3", "infra", "%3"),
        name="infra",
        state=ClaudeState.IDLE,
    )
    assert _matches(entry, "work:3")
    assert _matches(entry, "work")
    assert not _matches(entry, "main")


def test_empty_needle_matches_all():
    entry = _Entry(
        pane=make_pane("main", "0", "proj", "%0"),
        name="proj",
        state=ClaudeState.IDLE,
    )
    assert _matches(entry, "")


# ── Dashboard window exclusion logic ─────────────────────────────────────────

def test_claudre_window_excluded_from_popup():
    """The claudre dashboard window should not appear in the popup list."""
    panes = [
        make_pane("main", "0", "claudre", "%0"),  # dashboard — should be excluded
        make_pane("main", "1", "my-api",  "%1"),
    ]
    current_session = "main"
    current_idx = "0"

    filtered = [
        p for p in panes
        if not (
            p.session == current_session
            and p.window_index == current_idx
            and p.window_name == "claudre"
        )
    ]
    assert len(filtered) == 1
    assert filtered[0].window_name == "my-api"



"""Yes/No confirmation modal."""

from __future__ import annotations

from textual.screen import ModalScreen
from textual.containers import Vertical, Horizontal
from textual.widgets import Button, Label


class ConfirmScreen(ModalScreen[bool]):
    BINDINGS = [
        ("escape", "cancel", "Cancel"),
        ("y", "confirm_yes", "Yes"),
        ("n", "cancel", "No"),
    ]

    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = message

    def compose(self):
        with Vertical(id="confirm-dialog"):
            yield Label(self.message)
            with Horizontal():
                yield Button("Yes", variant="error", id="yes-btn")
                yield Button("No", variant="primary", id="no-btn")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss(event.button.id == "yes-btn")

    def action_confirm_yes(self) -> None:
        self.dismiss(True)

    def action_cancel(self) -> None:
        self.dismiss(False)



"""DataTable widget showing tmux window status."""

from __future__ import annotations

from textual.widgets import DataTable
from textual.widgets.data_table import ColumnKey

from claudre.models import ClaudeState, WindowState


def _state_markup(state: ClaudeState) -> str:
    match state:
        case ClaudeState.WORKING:
            return "[bold green]WORKING[/]"
        case ClaudeState.WAITING:
            return "[bold yellow]WAITING[/]"
        case ClaudeState.CRASHED:
            return "[bold red]CRASHED[/]"
        case ClaudeState.SUSPENDED:
            return "[dim]SUSPENDED[/]"
        case ClaudeState.IDLE:
            return "[dim]idle[/]"
        case ClaudeState.OPENING:
            return "[cyan]OPENING[/]"
        case _:
            return "[dim]unknown[/]"


def _row_cells(ws: WindowState) -> tuple[str, ...]:
    managed = "[green]●[/]" if ws.managed else "[dim]○[/]"
    branch = ws.vcs.branch or ""
    dirty = "[red]*[/]" if ws.vcs.dirty else ""
    state = _state_markup(ws.state)
    summary = ws.summary[:80] if ws.summary else "[dim]—[/]"
    if ws.summary_stale and ws.summary:
        summary = f"[dim]{summary}[/]"
    return (managed, ws.project_name, branch, dirty, state, summary)


# Column identifiers used with update_cell()
_COL_MANAGED  = "managed"
_COL_PROJECT  = "project"
_COL_BRANCH   = "branch"
_COL_DIRTY    = "dirty"
_COL_STATE    = "state"
_COL_SUMMARY  = "summary"

_COLUMNS: list[tuple[str, str]] = [
    (_COL_MANAGED,  "#"),
    (_COL_PROJECT,  "Project"),
    (_COL_BRANCH,   "Branch"),
    (_COL_DIRTY,    "M"),
    (_COL_STATE,    "State"),
    (_COL_SUMMARY,  "Summary"),
]


class WindowTable(DataTable):
    """Table displaying all tmux windows and their status."""

    BINDINGS: list = []
    _filter_text: str = ""

    def on_mount(self) -> None:
        self.cursor_type = "row"
        for col_key, label in _COLUMNS:
            self.add_column(label, key=col_key)

    # ------------------------------------------------------------------ #
    # Public API — incremental updates
    # ------------------------------------------------------------------ #

    def add_window(self, ws: WindowState) -> None:
        """Append a new row for the given window."""
        if self._filter_matches(ws):
            self.add_row(*_row_cells(ws), key=ws.pane_id)

    def update_window(self, ws: WindowState) -> None:
        """Update the row for an existing window in-place."""
        if not self._filter_matches(ws):
            # Row should not be visible — remove it if present
            try:
                self.remove_row(ws.pane_id)
            except Exception:
                pass
            return

        cells = _row_cells(ws)
        col_keys = [k for k, _ in _COLUMNS]
        try:
            for col_key, value in zip(col_keys, cells):
                self.update_cell(ws.pane_id, col_key, value)
        except Exception:
            # Row doesn't exist yet — add it
            try:
                self.add_row(*cells, key=ws.pane_id)
            except Exception:
                pass

    def remove_window(self, pane_id: str) -> None:
        """Remove the row for a given pane_id."""
        try:
            self.remove_row(pane_id)
        except Exception:
            pass

    def rebuild(self, windows: dict[str, WindowState]) -> None:
        """Full rebuild — preserves cursor position."""
        selected_key = self._selected_key()
        self.clear()
        restore_row = 0
        for i, ws in enumerate(windows.values()):
            if self._filter_matches(ws):
                self.add_row(*_row_cells(ws), key=ws.pane_id)
                if ws.pane_id == selected_key:
                    restore_row = i
        if self.row_count > 0:
            self.move_cursor(row=min(restore_row, self.row_count - 1))

    def get_selected_pane_id(self) -> str | None:
        return self._selected_key()

    def set_filter(self, text: str) -> None:
        self._filter_text = text.lower()

    # ------------------------------------------------------------------ #
    # Internal helpers
    # ------------------------------------------------------------------ #

    def _selected_key(self) -> str | None:
        if self.row_count == 0:
            return None
        try:
            return self.coordinate_to_cell_key(self.cursor_coordinate).row_key.value  # type: ignore[return-value]
        except Exception:
            return None

    def _filter_matches(self, ws: WindowState) -> bool:
        if not self._filter_text:
            return True
        needle = self._filter_text
        return (
            needle in ws.project_name.lower()
            or needle in ws.path.lower()
            or needle in ws.state.value.lower()
            or needle in ws.summary.lower()
        )



"""Shared data types for claudre v3."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Union


class ClaudeState(Enum):
    UNKNOWN = "unknown"
    IDLE = "idle"
    OPENING = "opening"
    WORKING = "working"
    WAITING = "waiting"
    SUSPENDED = "suspended"
    CRASHED = "crashed"


@dataclass
class TmuxPane:
    session: str
    window_index: str
    window_name: str
    pane_id: str
    pane_pid: int
    pane_command: str
    pane_path: str

    @property
    def target(self) -> str:
        return f"{self.session}:{self.window_index}"


@dataclass
class VcsStatus:
    branch: str = ""
    dirty: bool = False
    vcs_type: str | None = None


@dataclass
class WindowState:
    pane_id: str
    project_name: str
    path: str
    session: str = ""
    window_index: str = ""
    state: ClaudeState = ClaudeState.UNKNOWN
    vcs: VcsStatus = field(default_factory=VcsStatus)
    summary: str = ""
    summary_updated_at: datetime | None = None
    summary_stale: bool = False
    managed: bool = False  # @claudre_managed=1

    @property
    def target(self) -> str:
        """tmux window target, e.g. 'main:3'."""
        return f"{self.session}:{self.window_index}"


# Typed events for the event bus

@dataclass
class WindowStateChanged:
    pane_id: str
    old: ClaudeState
    new: ClaudeState


@dataclass
class SummaryUpdated:
    pane_id: str
    summary: str


@dataclass
class WindowAdded:
    pane_id: str


@dataclass
class WindowRemoved:
    pane_id: str


RegistryEvent = Union[WindowStateChanged, SummaryUpdated, WindowAdded, WindowRemoved]



"""Tests for templates.py."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from claudre.config import ClaudreConfig, TemplateConfig
from claudre.models import TmuxPane
from claudre.templates import resolve_template, _BUILTIN_TEMPLATES
from claudre.tmux_adapter import WindowSpec


def make_config(**kwargs) -> ClaudreConfig:
    return ClaudreConfig(**kwargs)


def test_resolve_builtin():
    config = make_config()
    tmpl = resolve_template("claude+terminal", config)
    assert tmpl.layout == "even-horizontal"
    assert "claude" in tmpl.pane_commands


def test_resolve_user_template():
    custom = TemplateConfig(layout="main-left", pane_commands=["nvim", "bash"])
    config = make_config(templates={"my-template": custom})
    tmpl = resolve_template("my-template", config)
    assert tmpl.layout == "main-left"
    assert tmpl.pane_commands == ["nvim", "bash"]


def test_resolve_unknown_falls_back():
    config = make_config()
    tmpl = resolve_template("nonexistent", config)
    assert tmpl == _BUILTIN_TEMPLATES["claude+terminal"]


def test_window_spec_fields():
    spec = WindowSpec(
        session="main",
        template_name="claude+terminal",
        project_name="myproj",
        start_directory="/home/user/myproj",
    )
    assert spec.session == "main"
    assert spec.template_name == "claude+terminal"
    assert spec.project_name == "myproj"


@pytest.mark.asyncio
async def test_create_from_template_single_pane():
    """create_from_template with a single-pane template marks managed and sends command."""
    config = make_config()
    spec = WindowSpec(
        session="main",
        template_name="claude",
        project_name="testproj",
        start_directory="/tmp/testproj",
    )

    primary = TmuxPane(
        session="main",
        window_index="2",
        window_name="testproj",
        pane_id="%5",
        pane_pid=999,
        pane_command="bash",
        pane_path="/tmp/testproj",
    )

    tmux = MagicMock()
    tmux.create_window = AsyncMock(return_value=primary)
    tmux.set_window_option = AsyncMock()
    tmux.rename_window = AsyncMock()
    tmux.select_layout = AsyncMock()
    tmux.split_window = AsyncMock()
    tmux.send_keys = AsyncMock()
    tmux.select_pane = AsyncMock()
    tmux.list_panes = AsyncMock(return_value=[primary])
    tmux.MANAGED_OPTION = "@claudre_managed"

    from claudre.templates import create_from_template
    result = await create_from_template(tmux, spec, config)

    assert result is primary
    # Should mark the window as managed
    tmux.set_window_option.assert_awaited_once_with(
        primary.target, "@claudre_managed", "1"
    )
    # Should rename the window
    tmux.rename_window.assert_awaited_once_with(primary.target, "testproj")
    # Should send the claude command to the single pane
    tmux.send_keys.assert_awaited_once()
    call_args = tmux.send_keys.call_args
    assert call_args[0][0] == "%5"
    assert "claude" in call_args[0][1]
    # Single-pane template — no splits
    tmux.split_window.assert_not_awaited()


@pytest.mark.asyncio
async def test_create_from_template_two_panes():
    """create_from_template with claude+terminal splits and sends to both panes."""
    config = make_config()
    spec = WindowSpec(
        session="main",
        template_name="claude+terminal",
        project_name="myproj",
        start_directory="/tmp/myproj",
    )

    pane1 = TmuxPane(
        session="main", window_index="3", window_name="myproj",
        pane_id="%10", pane_pid=100, pane_command="bash", pane_path="/tmp/myproj",
    )
    pane2 = TmuxPane(
        session="main", window_index="3", window_name="myproj",
        pane_id="%11", pane_pid=101, pane_command="bash", pane_path="/tmp/myproj",
    )

    tmux = MagicMock()
    tmux.create_window = AsyncMock(return_value=pane1)
    tmux.set_window_option = AsyncMock()
    tmux.rename_window = AsyncMock()
    tmux.select_layout = AsyncMock()
    tmux.split_window = AsyncMock()
    tmux.send_keys = AsyncMock()
    tmux.select_pane = AsyncMock()
    tmux.list_panes = AsyncMock(return_value=[pane1, pane2])
    tmux.MANAGED_OPTION = "@claudre_managed"

    from claudre.templates import create_from_template
    result = await create_from_template(tmux, spec, config)

    assert result is pane1
    # Layout should be applied twice (before and after splits)
    assert tmux.select_layout.await_count == 2
    # One split for the second pane
    tmux.split_window.assert_awaited_once()
    # Commands sent to both panes
    assert tmux.send_keys.await_count == 2
    # First pane focuses
    tmux.select_pane.assert_awaited_once_with("%10")



"""Tests for state_detector.py."""

from __future__ import annotations

import json
import time
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from claudre.models import ClaudeState, TmuxPane
from claudre.state_detector import JournalStateDetector


def make_pane(command: str = "claude", pid: int = 1234, path: str = "/tmp/proj") -> TmuxPane:
    return TmuxPane(
        session="main",
        window_index="0",
        window_name="test",
        pane_id="%1",
        pane_pid=pid,
        pane_command=command,
        pane_path=path,
    )


@pytest.fixture
def detector():
    return JournalStateDetector()


@pytest.mark.asyncio
async def test_detect_process_not_alive(detector):
    pane = make_pane(pid=999999)
    with patch.object(detector, "_is_process_alive", return_value=False):
        with patch.object(detector, "_get_sanitized_dir", return_value=None):
            state = await detector.detect(pane)
    assert state == ClaudeState.UNKNOWN


@pytest.mark.asyncio
async def test_detect_idle_command(detector):
    pane = make_pane(command="bash")
    with patch.object(detector, "_is_process_alive", return_value=True):
        state = await detector.detect(pane)
    assert state == ClaudeState.IDLE


@pytest.mark.asyncio
async def test_detect_opening_no_jsonl(detector):
    pane = make_pane(command="claude")
    with patch.object(detector, "_is_process_alive", return_value=True):
        with patch.object(detector, "_get_sanitized_dir", return_value=None):
            state = await detector.detect(pane)
    assert state == ClaudeState.OPENING


@pytest.mark.asyncio
async def test_detect_waiting(detector, tmp_path):
    # Create a JSONL file with last role == assistant, old mtime
    jsonl = tmp_path / "session.jsonl"
    jsonl.write_text(json.dumps({"role": "assistant", "content": "done"}) + "\n")
    # Make mtime old
    old_time = time.time() - 60
    import os
    os.utime(jsonl, (old_time, old_time))

    pane = make_pane(command="claude")
    with patch.object(detector, "_is_process_alive", return_value=True):
        with patch.object(detector, "_get_sanitized_dir", return_value="my_proj"):
        # The _find_latest_jsonl and _tail_read_last_role use real paths
        # so we patch them directly
            with patch.object(detector, "_find_latest_jsonl", return_value=jsonl):
                with patch.object(detector, "_get_cpu_percent", return_value=0.0):
                    state = await detector.detect(pane)
    assert state == ClaudeState.WAITING


@pytest.mark.asyncio
async def test_detect_crashed(detector, tmp_path):
    jsonl = tmp_path / "session.jsonl"
    jsonl.write_text(json.dumps({"role": "user", "content": "hello"}) + "\n")

    pane = make_pane(command="claude")
    with patch.object(detector, "_is_process_alive", return_value=False):
        with patch.object(detector, "_get_sanitized_dir", return_value="my_proj"):
            with patch.object(detector, "_find_latest_jsonl", return_value=jsonl):
                state = await detector.detect(pane)
    assert state == ClaudeState.CRASHED


@pytest.mark.asyncio
async def test_detect_suspended(detector, tmp_path):
    jsonl = tmp_path / "session.jsonl"
    jsonl.write_text(json.dumps({"role": "assistant", "content": "ok"}) + "\n")

    pane = make_pane(command="claude")
    with patch.object(detector, "_is_process_alive", return_value=False):
        with patch.object(detector, "_get_sanitized_dir", return_value="my_proj"):
            with patch.object(detector, "_find_latest_jsonl", return_value=jsonl):
                state = await detector.detect(pane)
    assert state == ClaudeState.SUSPENDED


def test_tail_read_last_role_empty_file(detector, tmp_path):
    empty = tmp_path / "empty.jsonl"
    empty.write_text("")
    assert detector._tail_read_last_role(empty) is None


def test_tail_read_last_role_finds_last(detector, tmp_path):
    jsonl = tmp_path / "session.jsonl"
    lines = [
        json.dumps({"role": "user", "content": "hello"}),
        json.dumps({"role": "assistant", "content": "hi"}),
        json.dumps({"role": "user", "content": "question"}),
    ]
    jsonl.write_text("\n".join(lines) + "\n")
    assert detector._tail_read_last_role(jsonl) == "user"


def test_build_path_cache_missing_dir(detector, monkeypatch, tmp_path):
    monkeypatch.setattr("claudre.state_detector.CLAUDE_DIR", tmp_path / "nonexistent")
    cache = detector._build_path_cache()
    assert cache == {}


def test_build_path_cache_reads_cwd(detector, monkeypatch, tmp_path):
    claude_dir = tmp_path / "projects"
    proj_dir = claude_dir / "-home-user-myproject"
    proj_dir.mkdir(parents=True)
    jsonl = proj_dir / "abc123.jsonl"
    jsonl.write_text(json.dumps({"cwd": "/home/user/myproject", "role": "user"}) + "\n")

    monkeypatch.setattr("claudre.state_detector.CLAUDE_DIR", claude_dir)
    cache = detector._build_path_cache()
    assert cache.get("/home/user/myproject") == "-home-user-myproject"


def test_get_jsonl_path_returns_none_for_unknown(detector, monkeypatch, tmp_path):
    monkeypatch.setattr("claudre.state_detector.CLAUDE_DIR", tmp_path / "nonexistent")
    result = detector.get_jsonl_path("/some/path")
    assert result is None


def test_get_jsonl_path_returns_latest(detector, monkeypatch, tmp_path):
    claude_dir = tmp_path / "projects"
    proj_dir = claude_dir / "-tmp-myproj"
    proj_dir.mkdir(parents=True)

    # Two JSONL files — latest should be returned
    old_jsonl = proj_dir / "old.jsonl"
    new_jsonl = proj_dir / "new.jsonl"
    old_jsonl.write_text(json.dumps({"cwd": "/tmp/myproj", "role": "user"}) + "\n")
    new_jsonl.write_text(json.dumps({"cwd": "/tmp/myproj", "role": "assistant"}) + "\n")

    import os, time as _time
    os.utime(old_jsonl, (_time.time() - 100, _time.time() - 100))
    os.utime(new_jsonl, (_time.time(), _time.time()))

    monkeypatch.setattr("claudre.state_detector.CLAUDE_DIR", claude_dir)
    result = detector.get_jsonl_path("/tmp/myproj")
    assert result == new_jsonl



"""Tests for config.py."""

import warnings

import pytest

from claudre.config import (
    ClaudreConfig,
    ConfigError,
    DefaultsConfig,
    ProjectConfig,
    TemplateConfig,
    load_config,
    _migrate_raw,
)


def test_defaults():
    cfg = ClaudreConfig()
    assert cfg.refresh_interval == 2.0
    assert cfg.summary_model == "claude-haiku-4-5-20251001"
    assert cfg.ai_summaries is True
    assert cfg.scope == "all"
    assert isinstance(cfg.defaults, DefaultsConfig)


def test_load_config_missing_file(tmp_path, monkeypatch):
    monkeypatch.setattr("claudre.config.CONFIG_PATH", tmp_path / "nonexistent.toml")
    cfg = load_config()
    assert isinstance(cfg, ClaudreConfig)


def test_load_config_valid(tmp_path, monkeypatch):
    toml = tmp_path / "config.toml"
    toml.write_text(
        """
refresh_interval = 5.0
ai_summaries = false

[defaults]
claude_command = "claude"
skip_permissions = true

[projects.myproj]
path = "/home/user/myproj"
quick_actions = ["make test", "make lint"]
"""
    )
    monkeypatch.setattr("claudre.config.CONFIG_PATH", toml)
    cfg = load_config()
    assert cfg.refresh_interval == 5.0
    assert cfg.ai_summaries is False
    assert cfg.defaults.skip_permissions is True
    assert "myproj" in cfg.projects
    assert cfg.projects["myproj"].quick_actions == ["make test", "make lint"]


def test_load_config_invalid_toml(tmp_path, monkeypatch):
    toml = tmp_path / "config.toml"
    toml.write_text("this is not valid toml ][}")
    monkeypatch.setattr("claudre.config.CONFIG_PATH", toml)
    with pytest.raises(ConfigError, match="parse error"):
        load_config()


def test_load_config_validation_error(tmp_path, monkeypatch):
    toml = tmp_path / "config.toml"
    toml.write_text("refresh_interval = 'not a number'")
    monkeypatch.setattr("claudre.config.CONFIG_PATH", toml)
    with pytest.raises(ConfigError, match="validation error"):
        load_config()


def test_migrate_dangerously_skip_permissions():
    raw = {"defaults": {"dangerously_skip_permissions": True}}
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        result = _migrate_raw(raw)
    assert result["defaults"].get("skip_permissions") is True
    assert "dangerously_skip_permissions" not in result["defaults"]
    assert any("deprecated" in str(warning.message).lower() for warning in w)


def test_migrate_project_layout():
    raw = {"projects": {"foo": {"path": "/tmp/foo", "layout": "even-horizontal"}}}
    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        result = _migrate_raw(raw)
    assert result["projects"]["foo"].get("template") == "even-horizontal"
    assert "layout" not in result["projects"]["foo"]


def test_template_config_defaults():
    t = TemplateConfig()
    assert t.layout == "even-horizontal"
    assert "claude" in t.pane_commands
    assert t.rename_to_project is True
